import numpy as np
from surpyval.nonparametric.nonparametric_fitter import NonParametricFitter
from .nelson_aalen import nelson_aalen as na
from .kaplan_meier import kaplan_meier as km
from .fleming_harrington import fleming_harrington as fh


def turnbull(x, c, n, t, estimator='Fleming-Harrington'):
    bounds = np.unique(np.concatenate([np.unique(x), np.unique(t)]))
    bounds = np.sort(np.concatenate([bounds, np.unique(x[c == 0])]))
    N = n.sum()

    if x.ndim == 1:
        x_new = np.empty(shape=(x.shape[0], 2))
        x_new[:, 0] = x
        x_new[:, 1] = x
        x = x_new

    # Unpack x array
    xl = x[:, 0]
    xr = x[:, 1]

    # Unpack t array
    tl = t[:, 0]
    tr = t[:, 1]

    # If there are left and right censored observations,
    # convert them to interval censored observations
    xl[c == -1] = -np.inf
    xr[c == 1] = np.inf

    # Find the count of intervals (M) and unique observation windows (N)
    M = bounds.size
    N = xl.size

    alpha = np.zeros(shape=(N, M))
    beta = np.ones(shape=(N, M))

    for i in range(0, N):
        x1, x2 = xl[i], xr[i]
        t1, t2 = tl[i], tr[i]
        if x1 == x2:
            # Find first index of repeated value
            idx = np.where(bounds == x1)[0][0]
            alpha[i, idx] = n[i]
        elif x2 == np.inf:
            alpha[i, :] = ((bounds > x1) & (bounds <= x2)).astype(int) * n[i]
            if x1 in x[c == 0]:
                idx = np.where(bounds == x1)[0][0]
                alpha[i, idx] = 0
        else:
            alpha[i, :] = ((bounds >= x1) & (bounds < x2)).astype(int) * n[i]
            if x1 in x[c == 0]:
                idx = np.where(bounds == x1)[0][0]
                alpha[i, idx] = 0

        beta[i, :] = (((bounds >= t1) & (bounds < t2)).astype(int))

    beta[:, M - 1] = 1
    n = n.reshape(-1, 1)
    d = np.zeros(M)
    p = np.ones(M) / M

    iters = 0
    p_prev = np.zeros_like(p)

    if estimator == 'Kaplan-Meier':
        func = km
    elif estimator == 'Nelson-Aalen':
        func = na
    else:
        func = fh

    old_err_state = np.seterr(all='ignore')

    while (iters < 1000) & (not np.allclose(p,
                                            p_prev,
                                            rtol=1e-30,
                                            atol=1e-30)):
        p_prev = p
        iters += 1
        ap = alpha * p
        # Observed deaths (left, right, or intervally 'observed')
        mu = alpha * ap / ap.sum(axis=1, keepdims=True)
        # Expected additional failures due to truncation
        nu = (n * (1 - beta) * (1 - beta) * p
              / (beta * p).sum(axis=1, keepdims=True))

        # Deaths/Failures
        d = (nu + mu).sum(axis=0)
        # M total observed and unobserved failures.
        M = (nu + mu).sum()
        # Risk set, i.e the number of items at risk at immediately before x
        r = M - d.cumsum() + d
        # Find the survival function values (R) using the deaths and risk set
        R = func(r, d)
        # Calculate the probability mass in each interval
        p = np.abs(np.diff(np.hstack([[1], R])))

        # The 'official' way to do it which is equivalent to using KM
        # p = (nu + mu).sum(axis=0)/(nu + mu).sum()

    out = {}
    out['x'] = bounds[1:-1]
    out['r'] = r[1:-1]
    out['d'] = d[1:-1]
    out['R'] = R[0:-2]
    out['R_upper'] = R[0:-2]
    out['R_lower'] = R[1:-1]
    out['alpha'] = alpha
    out['bounds'] = bounds

    np.seterr(**old_err_state)

    return out


class Turnbull_(NonParametricFitter):
    r"""
    Turnbull estimator class. Returns a `NonParametric` object from method
    :code:`fit()`. Calculates the Non-Parametric estimate of the survival
    function using the Turnbull NPMLE.

    Examples
    --------
    >>> import numpy as np
    >>> from surpyval import Turnbull
    >>> x = np.array([[1, 5], [2, 3], [3, 6], [1, 8], [9, 10]])
    >>> model = Turnbull.fit(x)
    >>> model.R
    array([1.        , 1.        , 0.63472351, 0.29479882, 0.2631432 ,
           0.2631432 , 0.2631432 , 0.09680497])
    """

    def __init__(self):
        self.how = 'Turnbull'


Turnbull = Turnbull_()



import numpy as np


def rank_adjust(t, c=None):
    """
    Currently limited to only Mean Order Number
    Room to expand to:
    Modal Order Number, and
    Median Order Number
    Uses mean order statistic to conduct rank adjustment
    For further reading see:
    http://reliawiki.org/index.php/Parameter_Estimation
    Above reference provides excellent explanation of how this method is
    derived this function currently assumes good input
    """
    # Total items in test/population
    N = len(t)
    # Preallocate adjusted ranks array
    ranks = np.zeros(N)

    if c is None:
        c = np.zeros(N)

    # Rank adjustment for [right] censored data
    # PMON - "Previous Mean Order Number"
    # NIPBSS - "Number of Items Before Present Suspended Set"
    PMON = 0
    for i in range(0, N):
        if c[i] == 0:
            NIBPSS = N - i
            ranks[i] = PMON + (N + 1 - PMON) / (1 + NIBPSS)
            PMON = ranks[i]
        elif c[i] == 1:
            ranks[i] = np.nan
        else:
            # ERROR
            raise ValueError("Censoring flag must be 0 or 1 with rank_adjust")
    return ranks



import pytest
import numpy as np
from surpyval import (
    Gumbel, Normal, Weibull, LogNormal,
    Logistic, LogLogistic, Uniform,
    Gamma, Beta, ExpoWeibull)

DISTS = [Gumbel, Normal, Weibull, LogNormal,
         Logistic, LogLogistic, Uniform,
         Beta, ExpoWeibull, Gamma]
parameter_sample_random_parameters = [((1, 20), (0.5, 5)),
                                  ((1, 100), (0.5, 100)),
                                  ((1, 100), (0.5, 20)),
                                  ((1, 3), (0.2, 1)),
                                  ((1, 100), (0.5, 20)),
                                  ((1, 100), (0.5, 20)),
                                  ((1, 100), (1, 100)),
                                  ((.1, 30), (.1, 30)),
                                  ((1, 30), (.1, 10), (0.5, 1.5)),
                                  ((1, 30), (.1, 10))]
FIT_SIZES = [1_000, 10_000, 100_000]


def generate_mle_test_cases():
    for idx, dist in enumerate(DISTS):
        random_parameters = parameter_sample_random_parameters[idx]
        for kind in ['full', 'censored', 'truncated', 'interval']:
            yield dist, random_parameters, kind


def generate_small_mle_test_cases():
    for idx, dist in enumerate(DISTS):
        random_parameters = parameter_sample_random_parameters[idx]
        for kind in ['full']:
            yield dist, random_parameters, kind


def generate_mpp_test_cases():
    for idx, dist in enumerate(DISTS):
        random_parameters = parameter_sample_random_parameters[idx]
        for rr in ['x', 'y']:
            yield dist, random_parameters, rr


def generate_mom_test_cases():
    for idx, dist in enumerate(DISTS):
        random_parameters = parameter_sample_random_parameters[idx]
        yield dist, random_parameters


def generate_mps_test_cases():
    for idx, dist in enumerate(DISTS):
        random_parameters = parameter_sample_random_parameters[idx]
        yield dist, random_parameters


def generate_mse_test_cases():
    for idx, dist in enumerate(DISTS):
        random_parameters = parameter_sample_random_parameters[idx]
        yield dist, random_parameters


def idfunc(x):
    if type(x) is tuple:
        return 'random_parameters'
    elif type(x) is str:
        return x
    else:
        return x.name


def interval_censor(x, n=100):
    n, xx = np.histogram(x, bins=n)
    x = np.vstack([xx[0:-1], xx[1::]]).T
    x = x[n > 0]
    n = n[n > 0]
    return x, n


def censor_at(x, q, where='right'):
    c = np.zeros_like(x)
    x = np.copy(x)
    if where == 'right':
        x_q = np.quantile(x, 1 - q)
        mask = x > x_q
        c[mask] = 1
        x[mask] = x_q
        return x, c
    elif where == 'left':
        x_q = np.quantile(x, q)
        mask = x < x_q
        c[mask] = -1
        x[mask] = x_q
        return x, c
    elif where == 'both':
        x_u = np.quantile(x, 1 - q)
        x_l = np.quantile(x, q)
        mask_l = x < x_l
        mask_u = x > x_u
        c[mask_l] = -1
        c[mask_u] = 1
        x[mask_l] = x_l
        x[mask_u] = x_u
        return x, c
    else:
        raise ValueError("'where' parameter not correctly defined")


def truncate_at(x, q, where='right'):
    x = np.copy(x)
    if where == 'right':
        x_q = np.quantile(x, 1 - q)
        x = x[x < x_q]
        return x, None, x_q
    elif where == 'left':
        x_q = np.quantile(x, q)
        x = x[x > x_q]
        return x, x_q, None
    elif where == 'both':
        x_u = np.quantile(x, 1 - q)
        x_l = np.quantile(x, q)
        x = x[x < x_u]
        x = x[x > x_l]
        return x, x_l, x_u
    else:
        raise ValueError("'where' parameter not correctly defined")


@pytest.mark.parametrize("dist,random_parameters,kind",
                         generate_mle_test_cases(),
                         ids=idfunc)
def test_mle_convergence(dist, random_parameters, kind):
    tol = 0.025
    for n in FIT_SIZES:
        test_params = []
        for b in random_parameters:
            test_params.append(np.random.uniform(*b))
        test_params = np.array(test_params)
        x = dist.random(n, *test_params)
        if kind == 'full':
            model = dist.fit(x)
        elif kind == 'censored':
            x, c = censor_at(x, 0.025, 'right')
            model = dist.fit(x, c=c)
        elif kind == 'truncated':
            x, tl, tr = truncate_at(x, 0.05, 'both')
            model = dist.fit(x, tl=tl, tr=tr)
        elif kind == 'interval':
            x, n = interval_censor(x)
            model = dist.fit(x=x, n=n)
        if model.params == []:
            continue
        fitted_params = np.array(model.params)
        max_params = np.max([fitted_params, test_params], axis=0)
        diff = np.abs(fitted_params - test_params) / max_params
        # Decrease the tolerance for every parameter
        # e.g. Weibull (2 params) tol will be 5%
        # ExpoWeibull the tolerance will be 7.5%
        if (diff < tol * dist.k).all():
            break
    else:
        raise AssertionError('MLE convergence not good for %s\n' % dist.name)


@pytest.mark.parametrize("dist,random_parameters,kind",
                         generate_small_mle_test_cases(),
                         ids=idfunc)
def test_mle_convergence_small(dist, random_parameters, kind):
    tol = 0.03
    for n in [100, 250, 500]:
        test_params = []
        for b in random_parameters:
            test_params.append(np.random.uniform(*b))
        test_params = np.array(test_params)
        x = dist.random(n, *test_params)
        if kind == 'full':
            model = dist.fit(x)
        if model.params == []:
            continue
        fitted_params = np.array(model.params)
        max_params = np.max([fitted_params, test_params], axis=0)
        diff = np.abs(fitted_params - test_params) / max_params
        # Decrease the tolerance for every parameter
        # e.g. Weibull (2 params) tol will be 6%
        # ExpoWeibull the tolerance will be 9%
        if (diff < tol * dist.k).all():
            break
    else:
        raise AssertionError('MLE fit for small data not good for %s\n'
                             % dist.name)


@pytest.mark.parametrize("dist,random_parameters,rr",
                         generate_mpp_test_cases(),
                         ids=idfunc)
def test_mpp(dist, random_parameters, rr):
    if dist not in [Beta, ExpoWeibull]:
        for n in FIT_SIZES:
            test_params = []
            tol = 0.025
            for b in random_parameters:
                test_params.append(np.random.uniform(*b))
            test_params = np.array(test_params)
            x = dist.random(10000, *test_params)
            model = dist.fit(x=x, rr=rr, how='MPP', heuristic='Nelson-Aalen')
            fitted_params = np.array(model.params)
            max_params = np.max([fitted_params, test_params], axis=0)
            diff = np.abs(fitted_params - test_params) / max_params
            if (diff < tol * dist.k).all():
                break
        else:
            raise AssertionError('MPP fit not very good in %s\n' % dist.name)


@pytest.mark.parametrize("dist,random_parameters",
                         generate_mom_test_cases(),
                         ids=idfunc)
def test_mom(dist, random_parameters):
    for n in FIT_SIZES:
        test_params = []
        tol = 0.025
        for b in random_parameters:
            test_params.append(np.random.uniform(*b))
        test_params = np.array(test_params)
        x = dist.random(n, *test_params)
        model = dist.fit(x=x, how='MOM')
        fitted_params = np.array(model.params)
        max_params = np.max([fitted_params, test_params], axis=0)
        diff = np.abs(fitted_params - test_params) / max_params
        if (diff < tol * dist.k).all():
            break
    else:
        raise AssertionError('MOM fit not very good in %s\n' % dist.name)


@pytest.mark.parametrize("dist,random_parameters",
                         generate_mps_test_cases(),
                         ids=idfunc)
def test_mps(dist, random_parameters):
    for n in FIT_SIZES:
        test_params = []
        tol = 0.01
        for b in random_parameters:
            test_params.append(np.random.uniform(*b))
        test_params = np.array(test_params)
        x = dist.random(n, *test_params)
        model = dist.fit(x=x, how='MPS')
        fitted_params = np.array(model.params)
        max_params = np.max([fitted_params, test_params], axis=0)
        diff = np.abs(fitted_params - test_params) / max_params
        if (diff < tol * dist.k).all():
            break
    else:
        raise AssertionError('MPS fit not very good in %s\n' % dist.name)


@pytest.mark.parametrize("dist,random_parameters",
                         generate_mse_test_cases(),
                         ids=idfunc)
def test_mse(dist, random_parameters):
    for n in FIT_SIZES:
        test_params = []
        # 5% accuracy!!
        tol = 0.05
        for b in random_parameters:
            test_params.append(np.random.uniform(*b))
        test_params = np.array(test_params)
        x = dist.random(n, *test_params)
        model = dist.fit(x=x, how='MSE')
        fitted_params = np.array(model.params)
        max_params = np.max([fitted_params, test_params], axis=0)
        diff = np.abs(fitted_params - test_params) / max_params
        if (diff < tol).all():
            break
    else:
        raise AssertionError('MPS fit not very good in %s\n' % dist.name)



from surpyval import np
from scipy.stats import uniform
from scipy.special import betaincinv
from scipy.special import gamma as gamma_func
from autograd.scipy.special import beta as abeta
from autograd_gamma import betainc as abetainc
from surpyval.parametric.parametric_fitter import ParametricFitter


class Beta_(ParametricFitter):
    def __init__(self, name):
        self.name = name
        self.k = 2
        self.bounds = ((0, None), (0, None),)
        self.support = (0, 1)
        self.plot_x_scale = 'linear'
        self.y_ticks = [0.0001, 0.0002, 0.0003, 0.001, 0.002,
                        0.003, 0.005, 0.01, 0.02, 0.03, 0.05,
                        0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8,
                        0.9, 0.95, 0.99, 0.999, 0.9999]
        self.param_names = ['alpha', 'beta']
        self.param_map = {
            'alpha': 0,
            'beta': 1
        }

    def _parameter_initialiser(self, x, c=None, n=None):
        if (c is not None) & ((c == 0).all()):
            x = np.repeat(x, n)
            p = self._mom(x)
        else:
            p = 1., 1.
        return p

    def sf(self, x, alpha, beta):
        r"""

        Survival (or reliability) function for the Beta Distribution:

        .. math::
            R(x) = 1 - \int_{0}^{x}t^{\alpha-1}\left (1 - t \right )^
            {\beta - 1}dt

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated
        alpha : numpy array or scalar
            One shape parameter for the Beta distribution
        beta : numpy array or scalar
            The scale parameter for the Beta distribution

        Returns
        -------

        df : scalar or numpy array
            The value(s) of the reliability function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Beta
        >>> x = np.array([.1, .2, .3, .4, .5])
        >>> Beta.sf(x, 3, 4)
        array([0.98415, 0.90112, 0.74431, 0.54432, 0.34375])
        """
        return 1 - self.ff(x, alpha, beta)

    def cs(self, x, X, alpha, beta):
        r"""

        Conditional survival (or reliability) function for the Beta
        Distribution:

        .. math::
            R(x, X) = \frac{R(x + X)}{R(X)}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated
        X : numpy array or scalar
            The value(s) at which each value(s) in x was known to have survived
        alpha : numpy array or scalar
            One shape parameter for the Beta distribution
        beta : numpy array or scalar
            The scale parameter for the Beta distribution

        Returns
        -------

        df : scalar or numpy array
            The value(s) of the conditional survival function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Beta
        >>> x = np.array([.1, .2, .3, .4, .5])
        >>> Beta.cs(x, 0.4, 3, 4)
        array([0.6315219 , 0.32921811, 0.12946429, 0.03115814, 0.00233319])
        """
        return self.sf(x + X, alpha, beta) / self.sf(X, alpha, beta)

    def ff(self, x, alpha, beta):
        r"""

        Failure (CDF or unreliability) function for the Beta Distribution:

        .. math::
            F(x) = \int_{0}^{x}t^{\alpha-1}\left (1 - t \right )^{\beta -1}dt

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated
        alpha : numpy array or scalar
            One shape parameter for the Beta distribution
        beta : numpy array or scalar
            The scale parameter for the Beta distribution

        Returns
        -------

        df : scalar or numpy array
            The value(s) of the failure function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Beta
        >>> x = np.array([.1, .2, .3, .4, .5])
        >>> Beta.ff(x, 3, 4)
        array([0.01585, 0.09888, 0.25569, 0.45568, 0.65625])
        """
        return abetainc(alpha, beta, x)

    def df(self, x, alpha, beta):
        r"""

        Density function for the Beta Distribution:

        .. math::
            f(x) = \frac{x^{\alpha-1}\left(1 - x \right )^{\beta-1}}{B \left (
            \alpha , \beta \right )}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated
        alpha : numpy array or scalar
            One shape parameter for the Beta distribution
        beta : numpy array or scalar
            The scale parameter for the Beta distribution

        Returns
        -------

        df : scalar or numpy array
            The value(s) of the density function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Beta
        >>> x = np.array([.1, .2, .3, .4, .5])
        >>> Beta.df(x, 3, 4)
        array([0.4374, 1.2288, 1.8522, 2.0736, 1.875 ])
        """
        return (x**(alpha - 1) * (1 - x)**(beta - 1)) / abeta(alpha, beta)

    def hf(self, x, alpha, beta):
        r"""

        Instantaneous hazard rate for the Beta distribution.

        .. math::
            h(x) = \frac{f(x)}{R(x)}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated
        alpha : numpy array or scalar
            One shape parameter for the Beta distribution
        beta : numpy array or scalar
            The scale parameter for the Beta distribution

        Returns
        -------

        df : scalar or numpy array
            The value(s) of the instantaneous hazard rate at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Beta
        >>> x = np.array([.1, .2, .3, .4, .5])
        >>> Beta.hf(x, 3, 4)
        array([0.44444444, 1.36363636, 2.48847926, 3.80952381, 5.45454545])
        """
        return self.df(x, alpha, beta) / self.sf(x, alpha, beta)

    def Hf(self, x, alpha, beta):
        r"""

        Cumulative hazard rate for the Beta distribution.

        .. math::
            H(x) = -\ln\left( R(x) \right )

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated
        alpha : numpy array or scalar
            One shape parameter for the Beta distribution
        beta : numpy array or scalar
            The scale parameter for the Beta distribution

        Returns
        -------

        df : scalar or numpy array
            The value(s) of the cumulative hazard function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Beta
        >>> x = np.array([.1, .2, .3, .4, .5])
        >>> Beta.hf(x, 3, 4)
        array([0.44444444, 1.36363636, 2.48847926, 3.80952381, 5.45454545])
        """
        return -np.log(self.sf(x, alpha, beta))

    def qf(self, p, alpha, beta):
        r"""

        Quantile function for the Beta Distribution:

        Parameters
        ----------

        p : numpy array or scalar
            The percentiles at which the quantile will be calculated
        alpha : numpy array or scalar
            One shape parameter for the Beta distribution
        beta : numpy array or scalar
            Another scale parameter for the Beta distribution

        Returns
        -------

        q : scalar or numpy array
            The quantiles for the Beta distribution at each value p.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Beta
        >>> p = np.array([.1, .2, .3, .4, .5])
        >>> Beta.qf(p, 3, 4)
        array([0.20090888, 0.26864915, 0.32332388, 0.37307973, 0.42140719])
        """
        return betaincinv(alpha, beta, p)

    def mean(self, alpha, beta):
        r"""

        Mean of the Beta distribution

        .. math::
            E = \frac{\alpha}{\alpha + \beta}

        Parameters
        ----------

        alpha : numpy array or scalar
            One shape parameter for the Beta distribution
        beta : numpy array or scalar
            Another scale parameter for the Beta distribution

        Returns
        -------

        mean : scalar or numpy array
            The mean(s) of the Beta distribution

        Examples
        --------
        >>> from surpyval import Beta
        >>> Beta.mean(3, 4)
        0.42857142857142855
        """
        return alpha / (alpha + beta)

    def moment(self, n, alpha, beta):
        r"""

        n-th (non central) moment of the Beta distribution

        .. math::
            E = \frac{\Gamma \left( n + \alpha\right )}{\beta^{n}\Gamma
            \left ( \alpha \right )}

        Parameters
        ----------

        n : integer or numpy array of integers
            The ordinal of the moment to calculate
        alpha : numpy array or scalar
            One shape parameter for the Beta distribution
        beta : numpy array or scalar
            Another scale parameter for the Beta distribution

        Returns
        -------

        moment : scalar or numpy array
            The moment(s) of the Beta distribution

        Examples
        --------
        >>> from surpyval import Beta
        >>> Beta.moment(2, 3, 4)
        0.75
        """
        return gamma_func(n + alpha) / (beta**n * gamma_func(alpha))

    def random(self, size, alpha, beta):
        r"""
        Draws random samples from the distribution in shape `size`

        Parameters
        ----------

        size : integer or tuple of positive integers
            Shape or size of the random draw
        alpha : numpy array or scalar
            One shape parameter for the Beta distribution
        beta : numpy array or scalar
            Another scale parameter for the Beta distribution

        Returns
        -------

        random : scalar or numpy array
            Random values drawn from the distribution in shape `size`

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Beta
        >>> Beta.random(10, 3, 4)
        array([0.48969376, 0.50266801, 0.66913503, 0.3031257 , 0.60897905,
               0.34901845, 0.34196432, 0.37401123, 0.06191741, 0.17604693])
        >>> Beta.random((5, 5), 3, 4)
        array([[0.68575237, 0.39409872, 0.18546004, 0.51266116, 0.55043579],
               [0.62656723, 0.69497065, 0.68958488, 0.37330801, 0.57053267],
               [0.07330674, 0.68020665, 0.42907148, 0.51884251, 0.12159803],
               [0.3566228 , 0.55493446, 0.59288881, 0.43542773, 0.31740851],
               [0.82044756, 0.23062323, 0.35342936, 0.2902573 , 0.54522114]])
        """
        U = uniform.rvs(size=size)
        return self.qf(U, alpha, beta)

    def mpp_y_transform(self, y, alpha, beta):
        return self.qf(y, alpha, beta)

    def mpp_inv_y_transform(self, y, alpha, beta):
        return abetainc(y, alpha, beta)

    def mpp_x_transform(self, x, gamma=0):
        return x - gamma

    def mpp(self, x, c=None, n=None, heuristic="Nelson-Aalen",
            rr='y', on_d_is_0=False, offset=False):
        msg = "Probability Plotting Method for Beta distribution"
        raise NotImplementedError(msg)

    def _mom(self, x):
        """
        MOM: Method of Moments for the beta distribution has an analytic answer
        """
        mean = x.mean()
        var = x.var()
        term1 = ((mean*(1 - mean)/var) - 1)
        alpha = term1 * mean
        beta = term1 * (1 - mean)

        return alpha, beta

    def var_R(self, dR, cv_matrix):
        dr_dalpha = dR[:, 0]
        dr_dbeta = dR[:, 1]
        var_r = (dr_dalpha**2 * cv_matrix[0, 0] +
                 dr_dbeta**2 * cv_matrix[1, 1] +
                 2 * dr_dalpha * dr_dbeta * cv_matrix[0, 1])
        return var_r


Beta = Beta_('Beta')



import numpy as np
from pandas import Series
from collections import defaultdict
from formulaic import Formula
import sys

import warnings

COX_PH_METHODS = ["breslow", "efron"]
FG_BASELINE_OPTIONS = [
    "Nelson-Aalen",
    "Kaplan-Meier"
]


def check_no_censoring(c):
    return any(c != 0)

def no_left_or_int(c):
    return any((c == -1) | (c == 2))

def surv_tolist(x):
    """
    supplement to the .tolist() function where there is mixed scalars and arrays.

    Survival data can be confusing sometimes. I'm probably not helping by trying
    to make it fit with one input.
    """
    if x.ndim == 2:        
        return [v[0] if v[0] == v[1] else v.tolist() for v in x]
    else:
        return x.tolist()

def check_and_convert_expected_float_array(a, name):
    if a is None:
        a = np.array([])
    else:
        try:
            a = np.array(a).astype(float, casting='safe')
        except Exception as e:
            raise ValueError("'{name}' must be an array of scalar numbers with real values.".format(name=name))
    return a

def group_xcn(x, c, n):
    """
    Ensures that all unique combinations of x and c are grouped. This should 
    reduce the chance of having arrays that are too long for the MLE fit.
    """

    # TODO: Change to numpy index
    grouped = defaultdict(lambda: defaultdict(int))
    if x.ndim == 2:
        for vx, vc, vn in zip(x, c, n):
            grouped[tuple(vx)][vc] += vn
    else:
        for vx, vc, vn in zip(x, c, n):
            grouped[vx][vc] += vn

    x_out = []
    c_out = []
    n_out = []

    for k, v in grouped.items():
        for cv, nv in v.items():
            x_out.append(k)
            c_out.append(cv)
            n_out.append(nv)
    return np.array(x_out), np.array(c_out), np.array(n_out)

def group_xcnt(x, c, n, t):
    """
    Ensures that all unique combinations of x and c are grouped. This should 
    reduce the chance of having arrays that are too long for the MLE fit.
    """
    grouped = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
    if x.ndim == 2:
        for vx, vc, vn, vt in zip(x, c, n, t):
            grouped[tuple(vx)][vc][tuple(vt)] += vn
    else:
        for vx, vc, vn, vt in zip(x, c, n, t):
            grouped[vx][vc][tuple(vt)] += vn

    x_out = []
    c_out = []
    n_out = []
    t_out = []

    for xv, level2 in grouped.items():
        for cv, level3 in level2.items():
            for tv, nv in level3.items():
                x_out.append(xv)
                c_out.append(cv)
                n_out.append(nv)
                t_out.append(tv)
    return np.array(x_out), np.array(c_out), np.array(n_out), np.array(t_out)

def xcn_sort(x, c, n):
    """
    Sorts the x, c, and n arrays so that x is increasing, and where there are ties they are ordered as left censored, failure, right censored, then intervally censored.

    Parameters
    ----------
    x: array
        array of values of variable for which observations were made.
    c: array
        array of censoring values (-1, 0, 1, 2) corrseponding to x
    n: array
        array of count of observations at each x and with censoring c

    Returns
    ----------
    x: array
        sorted array of values of variable for which observations were made.
    c: array
        sorted array of censoring values (-1, 0, 1, 2) corrseponding to rearranged x
    n: array
        array of count of observations at each rearranged x and with censoring c
    """
    
    idx_c = np.argsort(c, kind='stable')
    x = x[idx_c]
    c = c[idx_c]
    n = n[idx_c]

    if x.ndim == 1:
        idx = np.argsort(x, kind='stable')
    else:
        idx = np.argsort(x[:, 0], kind='stable')

    x = x[idx]
    c = c[idx]
    n = n[idx]
    return x, c, n

def xcn_handler(x, c=None, n=None):
    """
    Main handler that ensures any input to a surpyval fitter meets the requirements to be used in one of the parametric or nonparametric fitters.

    Parameters
    ----------
    x: array
        array of values of variable for which observations were made.
    c: array, optional (default: None)
        array of censoring values (-1, 0, 1, 2) corrseponding to x
    n: array, optional (default: None)
        array of count of observations at each x and with censoring c

    Returns
    ----------

    x: array
        sorted array of values of variable for which observations were made.
    c: array
        array of censoring values (-1, 0, 1, 2) corrseponding to output array x. If c was None, defaults to creating array of zeros the length of x.
    n: array
        array of count of observations at to output array x and with censoring c. If n was None, count array assumed to be all one observation.
    """
    if type(x) == list:
        if any([list == type(v) for v in x]):
            x_ndarray = np.empty(shape=(len(x), 2))
            for idx, val in enumerate(x):
                x_ndarray[idx, :] = np.array(val)
            x = x_ndarray
        else:
            x = np.array(x).astype(float)
    elif type(x) == Series:
        x = np.array(x)

    if x.ndim > 2:
        raise ValueError("Variable (x) array must be one or two dimensional")

    if x.ndim == 2:
        if x.shape[1] != 2:
            raise ValueError("Dimension 1 must be equalt to 2, try transposing data, or do you have a 1d array in a 2d array?")

        if not (x[:, 0] <= x[:, 1]).all():
            raise ValueError("All left intervals must be less than right intervals")

    if c is not None:
        c = np.array(c)
        if c.ndim != 1:
            raise ValueError("censoring array must be one dimensional")

        if c.shape[0] != x.shape[0]:
            raise ValueError("censoring array must be same length as variable array")

        if x.ndim == 2:
            if any(c[x[:, 0] == x[:, 1]] == 2):
                raise ValueError("Censor flag indicates interval censored but only has one failure time")

            if not all(c[x[:, 0] != x[:, 1]] == 2):
                raise ValueError("Censor flag provided, but case where interval flagged as non interval censoring")

            if any((c !=  0) & (c !=  1) & (c != -1) & (c !=  2)):
                raise ValueError("Censoring value must only be one of -1, 0, 1, or 2")

        else:
            if any((c !=  0) & (c !=  1) & (c != -1)):
                raise ValueError("Censoring value must only be one of -1, 0, 1 for single dimension input")
    else:
        c = np.zeros(x.shape[0])
        if x.ndim != 1:
            c[x[:, 0] != x[:, 1]] = 2

    if n is not None:
        n = np.array(n)
        if n.ndim != 1:
            raise ValueError("Count array must be one dimensional")
        if n.shape[0] != x.shape[0]:
            raise ValueError("count array must be same length as variable array.")
        if not (n > 0).all():
            raise ValueError("count array can't be 0")
    else:
        # Do check here for groupby and binning
        n = np.ones(x.shape[0])

    n = n.astype(int)
    c = c.astype(int)
    
    x, c, n = group_xcn(x, c, n)
    x, c, n = xcn_sort(x, c, n)

    return x, c, n

def xcnt_sort(x, c, n, t):
    """
    Sorts the x, c, and n arrays so that x is increasing, and where there are ties they are ordered as left censored, failure, right censored, then intervally censored.

    Parameters
    ----------
    x: array
        array of values of variable for which observations were made.
    c: array
        array of censoring values (-1, 0, 1, 2) corrseponding to x
    n: array
        array of count of observations at each x and with censoring c
    t: array, optional (default: None)
        array of values with shape (?, 2) with the left and right value of truncation   

    Returns
    ----------
    x: array
        sorted array of values of variable for which observations were made.
    c: array
        sorted array of censoring values (-1, 0, 1, 2) corrseponding to rearranged x
    n: array
        array of count of observations at each rearranged x and with censoring c
    t: array, optional (default: None)
        array of values with shape (?, 2) with the left and right value of truncation   
    """
    
    idx_c = np.argsort(c, kind='stable')
    x = x[idx_c]
    c = c[idx_c]
    n = n[idx_c]
    t = t[idx_c]

    if t.ndim == 1:
        idx = np.argsort(t, kind='stable')
    else:
        idx = np.argsort(t.min(axis=1), kind='stable')
    x = x[idx]
    c = c[idx]
    n = n[idx]
    t = t[idx]
    
    if x.ndim == 1:
        idx = np.argsort(x, kind='stable')
    else:
        idx = np.argsort(x.mean(axis=1), kind='stable')

    x = x[idx]
    c = c[idx]
    n = n[idx]
    t = t[idx]
    
    return x, c, n, t

def fsli_handler(f=None, s=None, l=None, i=None):
    """
    Converts the fsli format to the xcn format. This ensures is so that the 
    data can be passed to one of the parametric or nonparametric fitters.

    Parameters
    ----------
    f: array
        array of values for which the failure/death was observed
    s: array
        array of right censored observation values
    l: array
        array of left censored observation values
    i: array
        array of interval censored data

    Returns
    ----------
    f: array
        array of values for which the failure/death was observed that have
        been checked for correctness
    s: array
        array of right censored observation values that have been checked
        for correctness
    l: array
        array of left censored observation values that have been checked
        for correctness
    i: array
        array of interval censored data that have been checked for correctness
    """
    f = check_and_convert_expected_float_array(f, 'f')
    s = check_and_convert_expected_float_array(s, 's')
    l = check_and_convert_expected_float_array(l, 'l')
    i = check_and_convert_expected_float_array(i, 'i')

    if f.ndim != 1:
        raise ValueError("'f' array must be one-dimensional")
    if s.ndim != 1:
        raise ValueError("'s' array must be one-dimensional")
    if l.ndim != 1:
        raise ValueError("'l' array must be one-dimensional")

    if (i.ndim != 2)  & (i.size != 0):
        raise ValueError("'i' array must be one-dimensional")

    if (i.size != 0):
        if (i[:, 0] >= i[:, 1]).any():
            raise ValueError("Lower interval must not be greater than or equal to the upper interval")

    return f, s, l, i

def xrd_handler(x, r, d):
    """
    Takes a combination of 'x', 'r', and 'd' arrays and ensures that the data
    is feasible.

    Parameters
    ----------
    x: array
        array of values of variable for which observations were made.
    r: array
        array of at risk items at each value of x
    d: array
        array of failures / deaths at each value of x

    Returns
    ----------

    x: array
        array of values of variable for which observations were made.
    r: array
        array of at risk items at each value of x
    d: array
        array of failures / deaths at each value of x
    """

    try:
        x = np.array(x).astype(float)
    except Exception as e:
        raise ValueError("'x' must be an array of scalar numbers with real values.")

    try:
        r = np.array(r).astype(int, casting='safe')
    except Exception as e:
        raise ValueError("'r' must be an array of integers.")

    try:
        d = np.array(d).astype(int, casting='safe')
    except Exception as e:
        raise ValueError("'d' must be an array of integers.")

    if x.ndim != 1:
        raise ValueError("'x' must be a one dimensional array")

    if x.shape != r.shape:
        raise ValueError("'x' array not the same length as 'r' array")
    if x.shape != d.shape:
        raise ValueError("'x' array not the same length as 'd' array")

    if (d < 0).any():
        raise ValueError("'d' array cannot have any negative values")

    if (r <= 0).any():
        raise ValueError("'r' at risk item count array cannot have any negative values")

    if (d > r).any():
        raise ValueError("cannot have more deaths/failures than there are items at risk")

    return x, r, d


def xcnt_handler(x=None, c=None, n=None, t=None, 
                 xl=None, xr=None, tl=None, tr=None,
                 group_and_sort=True):
    """
    Main handler that ensures any input to a surpyval fitter meets the requirements to be used in one of the parametric or nonparametric fitters.

    Parameters
    ----------
    x: array
        array of values of variable for which observations were made.
    c: array, optional (default: None)
        array of censoring values (-1, 0, 1, 2) corrseponding to x
    n: array, optional (default: None)
        array of count of observations at each x and with censoring c
    t: array, optional (default: None)
        array of values with shape (?, 2) with the left and right value of truncation
    tl: array or scalar, optional (default: None)
        array of values of the left value of truncation. If scalar, all values will be treated as left truncated by that value
        cannot be used with 't' parameter but can be used with the 'tr' parameter
    tr: array or scalar, optional (default: None)
        array of values of the right value of truncation. If scalar, all values will be treated as right truncated by that value
        cannot be used with 't' parameter but can be used with the 'tl' parameter

    Returns
    ----------

    x: array
        sorted array of values of variable for which observations were made.
    c: array
        array of censoring values (-1, 0, 1, 2) corrseponding to output array x. If c was None, defaults to creating array of zeros the length of x.
    n: array
        array of count of observations at output array x and with censoring c. If n was None, count array assumed to be all one observation.
    t: array
        array of truncation values of observations at output array x and with censoring c.
    """

    if (x is None) & (xl is None) & (xr is None):
        raise ValueError("Must enter some data! Use either 'x' of both 'xl and 'xr'")

    if (x is not None) & ((xl is not None) | (xr is not None)):
        raise ValueError("Must use either 'x' of both 'xl and 'xr'")

    if (x is None) & ((xl is None) | (xr is None)):
        raise ValueError("Must use either 'x' of both 'xl and 'xr'")

    if x is None:
        xl = np.array(xl).astype(float)
        xr = np.array(xr).astype(float)
        x = np.vstack([xl, xr]).T

    if type(x) == list:
        if any([type(v) == list for v in x]):
            x_ndarray = np.empty(shape=(len(x), 2))
            for idx, val in enumerate(x):
                x_ndarray[idx, :] = np.array(val)
            x = x_ndarray
        else:
            x = np.array(x)
    elif type(x) == Series:
        x = np.array(x)

    if x.ndim > 2:
        raise ValueError("Variable (x) array must be one or two dimensional")

    if x.ndim == 2:
        if x.shape[1] != 2:
            raise ValueError("Dimension 1 must be equalt to 2, try transposing data, or do you have a 1d array in a 2d array?")

        if not (x[:, 0] <= x[:, 1]).all():
            raise ValueError("All left intervals must be less than (or equal to) right intervals")

    if c is not None:
        c = np.array(c)
        if c.ndim != 1:
            raise ValueError("Censoring flag array must be one dimensional")

        if c.shape[0] != x.shape[0]:
            raise ValueError("censoring flag array must be same length as variable array")

        if x.ndim == 2:
            if any(c[x[:, 0] == x[:, 1]] == 2):
                raise ValueError("Censor flag indicates interval censored but only has one failure time")

            if any((c == 2) & (x[:, 0] == x[:, 1])):
                raise ValueError("Censor flag provided, but case where interval flagged as non interval censoring")

            if any((c !=  0) & (c !=  1) & (c != -1) & (c !=  2)):
                raise ValueError("Censoring value must only be one of -1, 0, 1, or 2")

        else:
            if any((c !=  0) & (c !=  1) & (c != -1)):
                raise ValueError("Censoring value must only be one of -1, 0, 1 for single dimension input")

    else:
        c = np.zeros(x.shape[0])
        if x.ndim != 1:
            c[x[:, 0] != x[:, 1]] = 2

    if n is not None:
        n = np.array(n)
        if n.ndim != 1:
            raise ValueError("Count array must be one dimensional")
        if n.shape[0] != x.shape[0]:
            raise ValueError("count array must be same length as variable array.")
        if not (n > 0).all():
            raise ValueError("count array can't be 0 or less")
    else:
        # Do check here for groupby and binning
        n = np.ones(x.shape[0])

    if t is not None and ((tl is not None) or (tr is not None)):
        raise ValueError("Cannot use 't' with 'tl' or 'tr'. Use either 't' or any combination of 'tl' and 'tr'")

    elif (t is None) & (tl is None) & (tr is None):
        tl = np.ones(x.shape[0]) * -np.inf
        tr = np.ones(x.shape[0]) *  np.inf
        t  = np.vstack([tl, tr]).T
    elif (tl is not None) or (tr is not None):
        if tl is None:
            tl = np.ones(x.shape[0]) * -np.inf
        elif np.isscalar(tl):
            tl = np.ones(x.shape[0]) * tl
        else:
            tl = np.array(tl)
        
        if tr is None:
            tr = np.ones(x.shape[0]) * np.inf
        elif np.isscalar(tr):
            tr = np.ones(x.shape[0]) * tr
        else:
            tr = np.array(tr)
            
        if tl.ndim > 1:
            raise ValueError("Left truncation array must be one dimensional, did you mean to use 't'")
        if tr.ndim > 1:
            raise ValueError("Left truncation array must be one dimensional, did you mean to use 't'")
        if tl.shape[0] != x.shape[0]:
            raise ValueError("Left truncation array must be same length as variable array")
        if tr.shape[0] != x.shape[0]:
            raise ValueError("Right truncation array must be same length as variable array")
        if tl.shape != tr.shape:
            raise ValueError("Left truncation array and right truncation array must be the same length")
        t = np.vstack([tl, tr]).T

    else:
        t = np.array(t)
        if t.ndim != 2:
            raise ValueError("Truncation ndarray must be 2 dimensional")
        if t.shape[0] != x.shape[0]:
            raise ValueError("Truncation ndarray must be same shape as variable array")
            
    if (t[:, 1] <= t[:, 0]).any():
        raise ValueError("All left truncated values must be less than right truncated values")
    if x.ndim == 2:
        if ((t[:, 0] > x[:, 0]) & (np.isfinite(t[:, 0]))).any():
            raise ValueError("All left truncated values must be less than the respective observed values")
        elif ((t[:, 1] < x[:, 1]) & (np.isfinite(t[:, 1]))).any():            
            raise ValueError("All right truncated values must be greater than the respective observed values")
    else:
        if (t[:, 0] > x).any():            
            raise ValueError("All left truncated values must be less than the respective observed values")
        elif (t[:, 1] < x).any():            
            raise ValueError("All right truncated values must be greater than the respective observed values")

    x = x.astype(float)
    c = c.astype(int)
    n = n.astype(int)
    t = t.astype(float)
    
    if group_and_sort:
        x, c, n, t = group_xcnt(x, c, n, t)
        x, c, n, t = xcnt_sort(x, c, n, t)

    return x, c, n, t

def xcn_to_fsl(x, c=None, n=None):
    x = np.array(x)
    if c is None:
        c = np.zeros_like(x)
    if n is None:
        n = np.ones_like(x).astype(int)

    c = np.array(c)
    n = np.array(n).astype(int)

    f = np.repeat(x[c == 0], n[c == 0])
    s = np.repeat(x[c == 1], n[c == 1])
    l = np.repeat(x[c == -1], n[c == -1])
    return f, s, l

def xcn_to_fs(x, c=None, n=None):
    x = np.array(x)
    if c is None:
        c = np.zeros_like(x)
    if n is None:
        n = np.ones_like(x).astype(int)

    c = np.array(c)
    n = np.array(n).astype(int)

    f = np.repeat(x[c == 0], n[c == 0])
    s = np.repeat(x[c == 1], n[c == 1])
    return f, s

def xcnt_to_xrd(x, c=None, n=None, t=None, **kwargs):
    """
    Converts the xcn format to the xrd format.

    Parameters
    ----------
    x: array
        array of values of variable for which observations were made.
    c: array, optional (default: None)
        array of censoring values (-1, 0, 1, 2) corrseponding to x. If None, an array of 0s is created corresponding to each x.
    n: array, optional (default: None)
        array of count of observations at each x and with censoring c. If None, an array of ones is created.
    kwargs: keywords for truncation can be either 't' or a combo of 'tl' and 'tr'

    Returns
    ----------
    x: array
        sorted array of values of variable for which observations were made.
    r: array
        array of count of units/people at risk at time x.
    d: array
        array of the count of failures/deaths at each time x.
    """
    x, c, n, t = xcnt_handler(x, c, n, t, **kwargs)

    if np.isfinite(t[:, 1]).any():
        raise ValueError("xrd format can't be used right truncated data")
        
    if (t[:, 0] == t[0, 0]).all() & np.isfinite(t[0, 0]):
        print("Ignoring left truncated values as all observations truncated at same value", file=sys.stderr)
        
    if ((c != 1) & (c != 0)).any():
        raise ValueError("xrd format can't be used with left (c=-1) or interval (c=2) censoring")

    x = np.repeat(x, n)
    c = np.repeat(c, n)
    t = np.repeat(t[:, 0], n)
    n = np.ones_like(x).astype(int)

    x, idx = np.unique(x, return_inverse=True)

    d = np.bincount(idx, weights=1 - c)
    le = (t.reshape(-1, 1) <= x).sum(axis=0)
    # do is drop outs
    do = np.bincount(idx, weights=c)
    r = le + d - d.cumsum() + do - do.cumsum()
    r = r.astype(int)
    d = d.astype(int)
    return x, r, d

def xcn_to_xrd(x, c=None, n=None):
    """
    Converts the xcn format to the xrd format.

    Parameters
    ----------
    x: array
        array of values of variable for which observations were made.
    c: array, optional (default: None)
        array of censoring values (-1, 0, 1, 2) corrseponding to x. If None, an array of 0s is created corresponding to each x.
    n: array, optional (default: None)
        array of count of observations at each x and with censoring c. If None, an array of ones is created.

    Returns
    ----------
    x: array
        sorted array of values of variable for which observations were made.
    r: array
        array of count of units/people at risk at time x.
    d: array
        array of the count of failures/deaths at each time x.
    """
    x, c, n = xcn_handler(x, c, n)

    if ((c != 1) & (c != 0)).any():
        raise ValueError("xrd format can't handle left (c=-1) or interval (c=2) censoring")

    x = np.repeat(x, n)
    c = np.repeat(c, n)
    n = np.ones_like(x).astype(int)

    x, idx = np.unique(x, return_inverse=True)

    d = np.bincount(idx, weights=1 - c)
    # do is drop outs
    do = np.bincount(idx, weights=c)
    r = n.sum() + d - d.cumsum() + do - do.cumsum()
    r = r.astype(int)
    d = d.astype(int)
    return x, r, d

def xrd_to_xcn(x, r, d):
    """
    Exact inverse of the xcn_to_xrd function.
    ONLY USE IF YOU HAVE NO TRUNCATION
    """
    n_f = np.copy(d)
    x_f = np.copy(x)
    mask = n_f != 0
    n_f = n_f[mask]
    x_f = x_f[mask]

    delta = np.abs(np.diff(np.hstack([r, [0]])))

    sus = (delta - d)
    x_s = x[sus > 0]
    n_s = sus[sus > 0]

    x_f = np.repeat(x_f, n_f)
    x_s = np.repeat(x_s, n_s)

    return fs_to_xcn(x_f, x_s)

def fsli_to_xcn(f=None, s=None, l=None, i=None):
    """
    Converts the fsli format to the xcn format. This ensures is so that the 
    data can be passed to one of the parametric or nonparametric fitters.

    Parameters
    ----------
    f: array
        array of values for which the failure/death was observed
    s: array
        array of right censored observation values
    l: array
        array of left censored observation values
    i: array
        array of interval censored data

    Returns
    ----------
    x: array
        sorted array of values of variable for which observations were made.
    c: array
        array of censoring values (-1, 0, 1, 2) corrseponding to output array x.
    n: array
        array of count of observations at to output array x and with censoring c.
    """

    f, s, l, i = fsli_handler(f, s, l, i)
    x, c, n = fsl_to_xcn(f, s, l)

    if i.size == 0:
        return x, c, n
    else:
        x_i, n_i = np.unique(i, axis=0, return_counts=True)
        c_i = np.ones(x_i.shape[0]) * 2

        x_two = np.vstack([x, x]).T
        x = np.concatenate([x_two, x_i]).astype(float)
        c = np.hstack([c, c_i]).astype(int)
        n = np.hstack([n, n_i]).astype(int)

        x, c, n = xcn_sort(x, c, n)

        return x, c, n

def fsl_to_xcn(f=[], s=[], l=[]):
    x_f, n_f = np.unique(f, return_counts=True)
    c_f = np.zeros_like(x_f)

    x_s, n_s = np.unique(s, return_counts=True)
    c_s = np.ones_like(x_s)

    x_l, n_l = np.unique(l, return_counts=True)
    c_l = -np.ones_like(x_l)

    x = np.hstack([x_f, x_s, x_l])
    c = np.hstack([c_f, c_s, c_l]).astype(int)
    n = np.hstack([n_f, n_s, n_l]).astype(int)

    x, c, n = xcn_sort(x, c, n)

    return x, c, n

def fs_to_xcn(f=[], s=[]):
    return fsl_to_xcn(f, s, [])



def _scale(ll, n, scale):
    if scale:
        # Sometimes the ll can get too big. If we normalise it 
        # against the number of observations it is smaller
        # This allows for better performance of the optimizer.
        return ll.sum() / n.sum()
    else:
        return ll.sum()

def _get_idx(x_target, x):
    """
    Function to get the indices for a given vector of x values
    """
    x = np.atleast_1d(x)
    idx = np.argsort(x)
    rev = np.argsort(idx)
    x = x[idx]
    idx = np.searchsorted(x_target, x, side='right') - 1
    return idx, rev

def check_left_or_int_cens(c):
    if (-1 in c) or (2 in c):
        raise ValueError("Left or interval censoring not implemented with Competing Risks")

def check_Z_and_x(Z, x):
    if x.shape[0] != Z.shape[0]:
        raise ValueError("Z must have len(x) number of rows")

def check_e_and_x(e, x):
    if e.shape != x.shape:
        raise ValueError('Event vector, e, and duration vector, x, must have same shape')

def check_c_and_e(c, e):
    if     any(~(e[c == 1] == None)) \
        or any(~(e[c != 1] != None)):
        raise ValueError('None can only be used as event type for censored observation')

def wrangle_and_check_form_and_Z_cols(Z_cols, formula, df):
    if (Z_cols is None) & (formula is None):
        raise ValueError("'Z_cols' or 'formula' cannot both be None")
    
    if (Z_cols is not None) & (formula is not None):
        raise ValueError("Either 'Z_cols' or 'formula' must be provided; not both")
    
    if Z_cols is not None:
        unknown = [x for x in Z_cols if x not in df.columns]
        if len(unknown) > 0:
            raise ValueError("{} not in dataframe columns".format(unknown))
        Z = df[Z_cols].values.astype(float)
        mask = ~df[Z_cols].isna().any(axis=1).values
        Z = Z[mask]
        form = None
    else:
        form = Formula('0 + ' + formula)
        Z = form.get_model_matrix(df, na_action='ignore').values.astype(float)
        mask = ~np.any(np.isnan(Z), axis=1)

    return Z, mask, form

def wrangle_Z(Z):
    Z = np.array(Z)

    if Z.ndim == 1:
        Z = np.atleast_2d(Z).T
    elif Z.ndim == 2:
        pass
    else:
        raise ValueError("Covariate matrix must be two dimensional")

    mask = ~np.any(np.isnan(Z), axis=1)

    return Z[mask], mask

def validate_cr_df_inputs(df, x_col, e_col, 
                          c_col=None, n_col=None):
    x = df[x_col].values
    e = df[e_col].values

    if c_col:
        c = df[c_col].values
    else:
        c = None

    if n_col:
        n = df[n_col].values
    else:
        n = None
    return x, c, n, e

def validate_cr_inputs(x, c, n, e, method):
    """
    Validates the inputs prior to be used by the CoxPH model.
    """
    # Use existing surpyval validator. But don't group and sort
    # so as to put it out of order of the event array, e.
    x, c, n, _ = xcnt_handler(x, c, n, group_and_sort=False)

    e = np.array(e)
    (x, c, n) = (np.array(a).astype(float) for a in [x, c, n])

    # Check same shape
    check_e_and_x(e, x)

    # Not implemented, yet.
    if (-1 in c) or (2 in c):
        raise ValueError("Left or interval censoring not implemented with Competing Risks")

    # Ensure all cases where c is 0, e is not None and
    # where c is 1 e is None
    if     any(~(e[c == 1] == None)) \
        or any(~(e[c != 1] != None)):
        raise ValueError('None can only be used as event type for censored observation')

    # Two baselines
    # TODO: Add fh
    if method not in ['Nelson-Aalen', 'Kaplan-Meier']:
        raise ValueError("Unrecognised baseline method")

    return x, c, n, e

def validate_event(mapping, event):
    if event is not None and event not in mapping:
        raise ValueError("Event type not in model")

def validate_cif_event(event):
    if event is None:
        raise ValueError("CIF needs event type, not None")

def check_an_ids_tl_and_x(id, tl, x):
    """
    This function checks that, for a given item, id, the history
    of the item is complete. That is, the timeline provided in the
    steps from tl[0] > x[0] == tl[1] > x[1] == ... tl[-1] > x[-1]
    This ensures there are no overlaps nor are there any gaps.
    # TODO: Consider allowing gaps in timeline, but not overlaps!
    """
    if len(tl) != len(x):
        raise ValueError('Item {id} has differing lengths of the tl and x vectors'.format(id=id))
    n = len(x)
    if n != len(np.unique(tl)):
        raise ValueError('Item {id} has repeated tl values'.format(id=id))

    # Check finish times are unqiue
    if n != len(np.unique(x)):
        raise ValueError('Item {id} has repeated x values'.format(id=id))
    # Check all starts are less than stops
    if np.any(tl >= x):
        raise ValueError("tl has some values greater than or equal to x")

    # check that there is one start time and one finish time and
    # no other double ups or gaps
    x, n = np.unique([tl, x], return_counts=True)

    if n[0] != 1:
        raise ValueError("Multiple start times for item {id}".format(id=id))
    if n[-1] != 1:
        raise ValueError("Multiple end times for item {id}".format(id=id))

    if np.any(n[1:-1] != 2):
        raise ValueError("Missing or doubled-up time windows for item {id}".format(id=id))

def validate_tv_coxph(id, tl, x, Z, c, n):
    x, c, n, t = xcnt_handler(x, c, n, tl=tl, group_and_sort=False)

    if id is None:
        warnings.warn("No id provided, model fitted by coherence not checked")
    else:
        id = np.array(id)
        for i in id:
            tl_i = tl[id == i]
            x_i = x[id == i]
            check_an_ids_tl_and_x(i, tl_i, x_i)

    Z, mask = wrangle_Z(Z)
    x, c, n = [arr[mask] for arr in (x, c, n)]
    (x, c, n, Z) = (arr.astype(float) for arr in [x, c, n, Z])

    check_Z_and_x(Z, x)
    x, c, n, t = xcnt_handler(x, c, n, tl=tl, group_and_sort=False)

    return t[:, 0], x, Z, c, n

def validate_tv_coxph_df_inputs(df, id_col, tl_col, x_col,
                                Z_cols, c_col, n_col, formula):

    # TODO: Create count of dropped rows

    if x_col is None:
        raise ValueError("'x_col' not provided")
    elif x_col not in df.columns:
        raise ValueError("'{}' not in dataframe's columns".format(x_col))

    if tl_col is None:
        raise ValueError("'tl_col' not given")
    elif tl_col not in df.columns:
        raise ValueError("'{}' not in dataframe's columns".format(tl_col))

    if id_col is None:
        warnings.warn("'id_cols' is None, fit carried out without checking coherence of each id's timeline.")
        id = None
    elif id_col not in df.columns:
        raise ValueError("'{}' not in dataframe's columns".format(id_col))
    else:
        # Check to ensure each item (by id) has coherent data.
        # That is, make sure that there are no gaps or overlaps
        # in the timline of each item.
        id = df[id_col].values
        for i, s in df.groupby(id_col):
            tl = s[tl_col].values
            x = s[x_col].values
            check_an_ids_tl_and_x(i, tl, x)

    Z, mask, form = wrangle_and_check_form_and_Z_cols(Z_cols, formula, df)

    x = df.loc[mask, x_col].values
    tl = df.loc[mask, tl_col].values

    if c_col is None:
        c = None
    else:
        c = df.loc[mask, c_col].values

    if n_col is None:
        n = None
    else:
        n = df.loc[mask, n_col].values

    x, c, n, t = xcnt_handler(x, c, n, tl=tl, group_and_sort=False)

    return t[:, 0], x, Z, id, c, n, form

def validate_coxph_df_inputs(df, x_col, c_col, n_col, Z_cols, formula):
    # TODO: Return the count of dropped rows?

    Z, mask, form = wrangle_and_check_form_and_Z_cols(Z_cols, formula, df)
    
    x = df.loc[mask, x_col].values

    if c_col is None:
        c = None
    else:
        c = df.loc[mask, c_col].values

    if n_col is None:
        n = None
    else:
        n = df.loc[mask, n_col].values

    x, c, n, _ = xcnt_handler(x, c, n, group_and_sort=False)

    return x, c, n, Z, form

def validate_coxph(x, c, n, Z, tl, method):
    if method not in COX_PH_METHODS:
        raise ValueError("Method must be in {}".format(COX_PH_METHODS))

    x, c, n, t = xcnt_handler(x, c, n, tl=tl, group_and_sort=False)

    tl = t[:, 0]

    (x, c, n, Z, tl) = (np.array(a).astype(float) for a in [x, c, n, Z, tl])

    Z, mask = wrangle_Z(Z)
    x, c, n, tl = [arr[mask] for arr in (x, c, n, tl)]
    (x, c, n, tl, Z) = (arr.astype(float) for arr in [x, c, n, tl, Z])

    check_Z_and_x(Z, x)
    
    return x, c, n, tl, Z

def validate_fine_gray_inputs(x, Z, e, c, n):

    x, c, n, _ = xcnt_handler(x, c, n, group_and_sort=False)

    e = np.array(e)
    Z, mask = wrangle_Z(Z)
    x, c, n, e = [arr[mask] for arr in (x, c, n, e)]

    # Set all dtypes to float. Very poor results otherwise.
    (x, c, n, Z) = (arr.astype(float) for arr in [x, c, n, Z])

    check_e_and_x(e, x)
    check_Z_and_x(Z, x)
    check_c_and_e(c, e)
    check_left_or_int_cens(c)

    return x, Z, e, c, n



def tl_tr_to_t(tl=None, tr=None):
    if tl is not None: tl = np.array(tl)
    if tr is not None: tr = np.array(tr)
    
    if (tl is None) & (tr is None):
        raise ValueError("One input must not be None")
    elif (tl is not None) and (tr is not None):
        if tl.shape != tr.shape:
            raise ValueError("Left array must be the same shape as right array")
    elif tl is None:
        tl = np.ones_like(tr) * -np.inf
    elif tr is None:
        tr = np.ones_like(tl) *  np.inf

    if (tr < tl).any():
        raise ValueError("All left truncated values must be less than right truncated values")

    t = np.vstack([tl, tr]).T
    return t

def fs_to_xrd(f, s):
    """
    Chain of the fs_to_xrd and xrd_to_xcn functions.
    """
    x, c, n = fs_to_xcn(f, s)
    return xcn_to_xrd(x, c, n)


def fsl_to_xrd(f, s, l):
    """
    Chain of the fsl_to_xrd and xrd_to_xcn functions.
    """
    x, c, n = fsl_to_xcn(f, s, l)
    return xcn_to_xrd(x, c, n)


def round_sig(points, sig=2):
    """
    Used to round to sig significant figures.
    """
    places = sig - np.floor(np.log10(np.abs(points))) - 1
    output = []
    for p, i in zip(points, places):
        output.append(np.round(p, int(i)))
    return output



from autograd import numpy as np

class Linear_():
    def __init__(self):
        self.name = 'Linear'
        self.phi_param_map = {'a' : 0, 'b' : 1}
        self.phi_bounds = ((None, None), (None, None),)

    def phi(self, X, *params):
        a = params[0]
        b = params[1]
        return a + b * X

    def phi_init(self, life, X):
        X = X.flatten()
        b, a =  np.polyfit(X, life, 1)
        return [a, b]

Linear = Linear_()


import re
from autograd import jacobian
from autograd import grad
import autograd.numpy as np
from scipy.stats import uniform

from surpyval import round_sig, fsli_to_xcn
from scipy.special import ndtri as z
from surpyval import nonparametric as nonp
from copy import deepcopy, copy

import matplotlib.pyplot as plt
from scipy.optimize import approx_fprime

class Regression():
    """
    Result of ``.fit()`` or ``.from_params()`` method for every parametric surpyval 
    distribution.
    
    Instances of this class are very useful when a user needs the other functions 
    of a distribution for plotting, optimizations, monte carlo analysis and numeric 
    integration.

    """
    def __repr__(self):
        dist_params = self.params[0:self.k_dist]
        reg_model_params = self.params[self.k_dist:]
        dist_param_string = '\n'.join(['{:>10}'.format(name) + ": " 
            + str(p) for p, name in zip(dist_params, self.distribution.param_names) 
            if name not in self.fixed])

        reg_model_param_string = '\n'.join(['{:>10}'.format(name) + ": " 
            + str(p) for p, name in zip(reg_model_params, self.reg_model.phi_param_map) 
            if name not in self.fixed])

        if hasattr(self, 'params'):
            out = ('Parametric Regression SurPyval Model'
                   + '\n===================================='
                   + '\nKind                : {kind}'
                   + '\nDistribution        : {dist}'
                   + '\nRegression Model    : {reg_model}'
                   + '\nFitted by           : MLE'
                   ).format(kind=self.kind, 
                            dist=self.distribution.name, 
                            reg_model=self.reg_model.name)

            out = (out + '\nDistribution        :\n'
                       + '{params}'.format(params=dist_param_string))

            out = (out + '\nRegression Model    :\n'
                       + '{params}'.format(params=reg_model_param_string))

            return out
        else:
            return "Unable to fit values"

    def phi(self, X):
        return self.reg_model.phi(X, *self.phi_params)

    def sf(self, x, X):
        r"""
        Surival (or Reliability) function for a distribution using the parameters found in the ``.params`` attribute.

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the survival function will be calculated 

        Returns
        -------

        sf : scalar or numpy array 
            The scalar value of the survival function of the distribution if a scalar was passed. If an array like object was passed then a numpy array is returned with the value of the survival function at each corresponding value in the input array.


        Examples
        --------
        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> model.sf(2)
        0.9920319148370607
        >>> model.sf([1, 2, 3, 4, 5])
        array([0.9990005 , 0.99203191, 0.97336124, 0.938005  , 0.8824969 ])
        """
        if type(x) == list:
            x = np.array(x)
        return self.model.sf(x, X, *self.params)

    def ff(self, x, X):
        r"""
        The cumulative distribution function, or failure function, for a distribution using the parameters found in the ``.params`` attribute.

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the failure function (CDF) will be calculated

        Returns
        -------

        ff : scalar or numpy array 
            The scalar value of the CDF of the distribution if a scalar was passed. If an array like object was passed then a numpy array is returned with the value of the CDF at each corresponding value in the input array.


        Examples
        --------

        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> model.ff(2)
        0.007968085162939342
        >>> model.ff([1, 2, 3, 4, 5])
        array([0.0009995 , 0.00796809, 0.02663876, 0.061995  , 0.1175031 ])
        """
        if type(x) == list:
            x = np.array(x)

        return self.model.ff(x, X, *self.params)

    def df(self, x, X):
        r"""
        The density function for a distribution using the parameters found in the ``.params`` attribute.

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the density function will be calculated

        Returns
        -------

        df : scalar or numpy array 
            The scalar value of the density function of the distribution if a scalar was passed. If an array like object was passed then a numpy array is returned with the value of the density function at each corresponding value in the input array.


        Examples
        --------

        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> model.df(2)
        0.01190438297804473
        >>> model.df([1, 2, 3, 4, 5])
        array([0.002997  , 0.01190438, 0.02628075, 0.04502424, 0.06618727])
        """
        if type(x) == list:
            x = np.array(x)
        return self.model.df(x, X, *self.params)

    def hf(self, x, X):
        r"""
        The instantaneous hazard function for a distribution using the parameters found in the ``.params`` attribute.

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the instantaneous hazard function will be calculated

        Returns
        -------

        hf : scalar or numpy array 
            The scalar value of the instantaneous hazard function of the distribution if a scalar was passed. If an array like object was passed then a numpy array is returned with the value of the instantaneous hazard function at each corresponding value in the input array.


        Examples
        --------

        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> model.hf(2)
        0.012000000000000002
        >>> model.hf([1, 2, 3, 4, 5])
        array([0.003, 0.012, 0.027, 0.048, 0.075])
        """
        if type(x) == list:
            x = np.array(x)
        return self.model.hf(x, X, *self.params)

    def Hf(self, x, X):
        r"""

        The cumulative hazard function for a distribution using the parameters found in the ``.params`` attribute.

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the cumulative hazard function will be calculated

        Returns
        -------

        Hf : scalar or numpy array 
            The scalar value of the cumulative hazard function of the distribution if a scalar was passed. If an array like object was passed then a numpy array is returned with the value of the cumulative hazard function at each corresponding value in the input array.


        Examples
        --------

        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> model.Hf(2)
        0.008000000000000002
        >>> model.Hf([1, 2, 3, 4, 5])
        array([0.001, 0.008, 0.027, 0.064, 0.125])
        """
        if type(x) == list:
            x = np.array(x)
        return self.model.hf(x, X, *self.params)

    def random(self, size, X):
        r"""

        A method to draw random samples from the distributions using the parameters found in the ``.params`` attribute.

        Parameters
        ----------
        size : int
            The number of random samples to be drawn from the distribution.

        X : scalar or array like
            The value(s) of the stresses at which the random 

        Returns
        -------
        random : numpy array 
            Returns a numpy array of size ``size`` with random values drawn from the distribution.


        Examples
        --------
        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> np.random.seed(1)
        >>> model.random(1)
        array([8.14127103])
        >>> model.random(10)
        array([10.84103403,  0.48542084,  7.11387062,  5.41420125,  4.59286657,
                5.90703589,  7.5124326 ,  7.96575225,  9.18134126,  8.16000438])
        """
        if (self.p == 1) and (self.f0 == 0):
            return self.dist.qf(uniform.rvs(size=size), *self.params) + self.gamma
        elif (self.p != 1) and (self.f0 == 0):
            n_obs = np.random.binomial(size, self.p)

            f = self.dist.qf(uniform.rvs(size=n_obs), *self.params) + self.gamma
            s = np.ones(np.array(size) - n_obs) * np.max(f) + 1

            return fsli_to_xcn(f, s)

        elif (self.p == 1) and (self.f0 != 0):
            n_doa = np.random.binomial(size, self.f0)

            x0 = np.zeros(n_doa) + self.gamma
            x = self.dist.qf(uniform.rvs(size=size - n_doa), *self.params) + self.gamma
            x = np.concatenate([x, x0])
            np.random.shuffle(x)

            return x
        else:
            N = np.random.multinomial(1, [self.f0, self.p - self.f0, 1. - self.p], size).sum(axis=0)
            N = np.atleast_2d(N)
            n_doa, n_obs, n_cens = N[:, 0], N[:, 1], N[:, 2]
            x0 = np.zeros(n_doa) + self.gamma
            x = self.dist.qf(uniform.rvs(size=n_obs), *self.params) + self.gamma
            f = np.concatenate([x, x0])
            s = np.ones(n_cens) * np.max(f) + 1
            # raise NotImplementedError("Combo zero-inflated and lfp model not yet supported")
            return fsli_to_xcn(f, s)

    def neg_ll(self):
        r"""

        The the negative log-likelihood for the model, if it was fit with the ``fit()`` method. Not available if fit with the ``from_params()`` method.

        Parameters
        ----------

        None

        Returns
        -------

        neg_ll : float
            The negative log-likelihood of the model

        Examples
        --------

        >>> from surpyval import Weibull
        >>> import numpy as np
        >>> np.random.seed(1)
        >>> x = Weibull.random(100, 10, 3)
        >>> model = Weibull.fit(x)
        >>> model.neg_ll()
        262.52685642385734
        """
        if not hasattr(self, 'data'):
            raise ValueError("Must have been fit with data")

        return self._neg_ll

    def bic(self):
        r"""

        The the Bayesian Information Criterion (BIC) for the model, if it was fit with the ``fit()`` method. Not available if fit with the ``from_params()`` method.

        Parameters
        ----------

        None

        Returns
        -------

        bic : float
            The BIC of the model

        Examples
        --------

        >>> from surpyval import Weibull
        >>> import numpy as np
        >>> np.random.seed(1)
        >>> x = Weibull.random(100, 10, 3)
        >>> model = Weibull.fit(x)
        >>> model.bic()
        534.2640532196908

        References:
        -----------

        `Bayesian Information Criterion for Censored Survival Models <https://www.jstor.org/stable/2677130>`_.

        """
        if hasattr(self, '_bic'):
            return self._bic
        else:
            self._bic = self.k  * np.log(self.data['n'][self.data['c'] == 0].sum()) + 2 * self.neg_ll()
            return self._bic

    def aic(self): 
        r"""

        The the Aikake Information Criterion (AIC) for the model, if it was fit with the ``fit()`` method. Not available if fit with the ``from_params()`` method.

        Parameters
        ----------

        None

        Returns
        -------

        aic : float
            The AIC of the model

        Examples
        --------

        >>> from surpyval import Weibull
        >>> import numpy as np
        >>> np.random.seed(1)
        >>> x = Weibull.random(100, 10, 3)
        >>> model = Weibull.fit(x)
        >>> model.aic()
        529.0537128477147
        """
        if hasattr(self, '_aic'):
            return self._aic
        else:
            self._aic = 2 * self.k + 2 * self.neg_ll()
            return self._aic

    def aic_c(self):
        r"""

        The the Corrected Aikake Information Criterion (AIC) for the model, if it was fit with the ``fit()`` method. Not available if fit with the ``from_params()`` method.

        Parameters
        ----------

        None

        Returns
        -------

        aic_c : float
            The Corrected AIC of the model

        Examples
        --------

        >>> from surpyval import Weibull
        >>> import numpy as np
        >>> np.random.seed(1)
        >>> x = Weibull.random(100, 10, 3)
        >>> model = Weibull.fit(x)
        >>> model.aic()
        529.1774241879209
        """
        if hasattr(self, '_aic_c'):
            return self._aic_c
        else:
            k = len(self.params)
            n = self.data['n'].sum()
            self._aic_c = self.aic() + (2*k**2 + 2*k)/(n - k - 1)
            return self._aic_c



import autograd.numpy as np
from autograd import jacobian, hessian, elementwise_grad
from scipy.optimize import minimize
import surpyval
import inspect

from ..parametric.fitters import bounds_convert, fix_idx_and_function
from .regression import Regression

class AcceleratedFailureTimeFitter():
    def __init__(self, name, distribution, acc_model):

        if str(inspect.signature(acc_model.phi)) != '(X, *params)':
            raise ValueError('PH function must have the signature \'(X, *params)\'')

        self.name = name
        self.dist = distribution
        self.acc_model = acc_model
        self.k_dist = len(self.dist.param_names)
        self.bounds = self.dist.bounds
        self.support = self.dist.support
        self.param_names = self.dist.param_names
        self.param_map = {v : i for i, v in enumerate(self.dist.param_names)}
        self.phi = acc_model.phi
        self.Hf_dist = self.dist.Hf
        self.hf_dist = self.dist.hf
        self.sf_dist = self.dist.sf
        self.ff_dist = self.dist.ff
        self.df_dist = self.dist.df

    def Hf(self, x, X, *params):
        dist_params = np.array(params[0:self.k_dist])
        phi_params = np.array(params[self.k_dist:])
        return self.Hf_dist(self.phi(X, *phi_params) * x, *dist_params)

    def hf(self, x, X, *params):
        dist_params = np.array(params[0:self.k_dist])
        phi_params = np.array(params[self.k_dist:])
        return self.hf_dist(self.phi(X, *phi_params) * x, *dist_params)

    def df(self, x, X, *params):
        return self.hf(x, X, *params) * np.exp(-self.Hf(x, X, *params))

    def sf(self, x, X, *params):
        return np.exp(-self.Hf(x, X, *params))

    def ff(self, x, X, *params):
        return 1 - np.exp(-self.Hf(x, X, *params))

    def _parameter_initialiser_dist(self, x, c=None, n=None, t=None):
        out = []
        for low, high in self.bounds:
            if (low is None) & (high is None):
                out.append(0)
            elif high is None:
                out.append(low + 1.)
            elif low is None:
                out.append(high - 1.)
            else:
                out.append((high + low)/2.)

        return out

    def mpp_inv_y_transform(self, y, *params):
        return y

    def mpp_y_transform(self, y, *params):
        return y

    def mpp_inv_y_transform(self, y, *params):
        return y

    def mpp_x_transform(self, x, gamma=0):
        return x - gamma

    def log_df(self, x, X, *params):
        return np.log(self.hf(x, X, *params)) - self.Hf(x, X, *params)

    def log_sf(self, x, X, *params):
        return -self.Hf(x, X, *params)

    def log_ff(self, x, X, *params):
        return np.log(self.ff(x, X, *params))

    def neg_ll(self, X, x, c, n, *params):
        params = np.array(params)

        like = np.zeros_like(x).astype(float)
        like = np.where(c ==  0, self.log_df(x, X, *params), like)
        like = np.where(c ==  1, self.log_sf(x, X, *params), like)
        like = np.where(c ==  -1, self.log_ff(x, X, *params), like)

        like = np.multiply(n, like)
        return -np.sum(like)

    def random(self, size, X, *params):
        dist_params = np.array(params[0:self.k_dist])
        phi_params = np.array(params[self.k_dist:])

        x = []
        X_out = []

        for stress in np.unique(X):
            U = np.random.uniform(0, 1, size)
            x.append(self.dist.qf(U, *dist_params)/self.phi(stress, *phi_params))
            X_out.append(np.ones(size) * stress)
        return np.concatenate(x), np.concatenate(X_out)

    def fit(self, X, x, c=None, n=None, t=None, init=[], fixed={}):
        x, c, n, t = surpyval.xcnt_handler(x=x, c=c, n=n, t=t, group_and_sort=False)

        if init == []:
            stress_data = np.unique(X, axis=0)
            params_at_X = []
            for s in stress_data:
                params_at_X.append(self.dist.fit(x[X == s], c[X == s], n[X == s]).params)

            params_at_X = np.array(params_at_X)
            dist_init = params_at_X.mean(axis=0)

            acc_parameter_data = params_at_X[:, self.param_map[self.fixed_parameter]]
            acc_parameter_data = self.acc_parameter_relationship(acc_parameter_data)

            if callable(self.acc_model.phi_init):
                phi_init = self.acc_model.phi_init(acc_parameter_data, stress_data)
            else:
                phi_init = self.acc_model.phi_init


            init = np.array([*dist_init, *phi_init])
        else:
            init = np.array(init)

        if self.fixed != {}:
            fixed = {**self.fixed, **fixed}

        # Dynamic or static bounds determination
        if callable(self.acc_model.phi_bounds):
            bounds = (*self.bounds, *self.acc_model.phi_bounds(X))
        else:
            bounds = (*self.bounds, *self.acc_model.phi_bounds)

        if callable(self.acc_model.phi_param_map):
            phi_param_map = self.acc_model.phi_param_map(X)
        else:
            phi_param_map = self.acc_model.phi_param_map

        param_map = {**self.param_map, **{k : v + len(self.param_map) for k, v in phi_param_map.items()}}

        transform, inv_trans, funcs, inv_f = bounds_convert(x, bounds)
        const, fixed_idx, not_fixed = fix_idx_and_function(fixed, param_map, funcs)

        init = transform(init)[not_fixed]

        with np.errstate(all='ignore'):
            fun  = lambda params : self.neg_ll(X, x, c, n, *inv_trans(const(params)))
            # fun  = lambda params : self.neg_ll(X, x, c, n, *params)
            # jac = jacobian(fun)
            # hess = hessian(fun)
            res = minimize(fun, init)
            res = minimize(fun, res.x, method='TNC')
            # res = minimize(fun, init, jac=jac, method='BFGS')
            # res = minimize(fun, init, method='Newton-CG', jac=jac)

        params = inv_trans(const(res.x))
        model = Regression()
        model.model = self
        model.reg_model = self.acc_model
        model.kind = "Accelerated Failure Time"
        model.distribution = self.dist
        model.params = np.array(params)
        model.res = res
        model._neg_ll = res['fun']
        model.fixed = self.fixed
        model.k_dist = self.k_dist
        model.k = len(bounds)

        model.data = {
            'x' : x,
            'c' : c,
            'n' : n,
            't' : t
        }

        return model



from surpyval import np
from scipy.stats import uniform
from scipy.special import ndtri as z
from autograd.scipy.stats import norm
from scipy.stats import norm as scipy_norm
from surpyval import nonparametric as nonp
from surpyval import parametric as para
from surpyval.parametric.parametric_fitter import ParametricFitter

class LogNormal_(ParametricFitter):
    def __init__(self, name):
        self.name = name
        self.k = 2
        self.bounds = ((0, None), (0, None),)
        self.support = (0, np.inf)
        self.plot_x_scale = 'log'
        self.y_ticks = [0.001, 0.01, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.99, 0.999]
        self.param_names = ['mu', 'sigma']
        self.param_map = {
            'mu'    : 0,
            'sigma' : 1
        }

    def _parameter_initialiser(self, x, c=None, n=None, t=None, offset=False):
        # Need an offset mpp function here
        norm_mod = para.Normal.fit(np.log(x), c=c, n=n, t=t, how='MLE')
        mu, sigma = norm_mod.params
        return mu, sigma

    def sf(self, x, mu, sigma):
        r"""

        Surival (or Reliability) function for the LogNormal Distribution:

        .. math::
            R(x) = 1 - \Phi \left( \frac{\ln(x) - \mu}{\sigma} \right )

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the LogNormal distribution
        sigma : numpy array or scalar
            The scale parameter for the LogNormal distribution

        Returns
        -------

        sf : scalar or numpy array 
            The value(s) of the survival function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogNormal
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> LogNormal.sf(x, 3, 4)
        array([0.77337265, 0.71793339, 0.68273014, 0.65668272, 0.63594491])
        """
        return 1 - self.ff(x, mu, sigma)

    def cs(self, x, X, mu, sigma):
        r"""

        Conditional survival function for the LogNormal Distribution:

        .. math::
            R(x, X) = \frac{R(x + X)}{R(X)}

        Parameters
        ----------

        x : numpy array or scalar
            The value(s) at which the function will be calculated
        X : numpy array or scalar
            The value(s) at which each value(s) in x was known to have survived
        mu : numpy array or scalar
            The location parameter for the LogNormal distribution
        sigma : numpy array or scalar
            The scale parameter for the LogNormal distribution

        Returns
        -------

        cs : scalar or numpy array 
            the conditional survival probability at x

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogNormal
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> LogNormal.cs(x, 5, 3, 4)
        array([0.97287811, 0.9496515 , 0.92933892, 0.91129122, 0.89505592])
        """
        return self.sf(x + X, mu, sigma) / self.sf(X, mu, sigma)

    def ff(self, x, mu, sigma):
        r"""

        Failure (CDF or unreliability) function for the LogNormal Distribution:

        .. math::
            F(x) = \Phi \left( \frac{\ln(x) - \mu}{\sigma} \right )

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the LogNormal distribution
        sigma : numpy array or scalar
            The scale parameter for the LogNormal distribution

        Returns
        -------

        ff : scalar or numpy array 
            The value(s) of the failure function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogNormal
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> LogNormal.ff(x, 3, 4)
        array([0.22662735, 0.28206661, 0.31726986, 0.34331728, 0.36405509])
        """
        return norm.cdf(np.log(x), mu, sigma)

    def df(self, x, mu, sigma):
        r"""

        Density function for the LogNormal Distribution:

        .. math::
            f(x) = \frac{1}{x \sigma \sqrt{2\pi}}e^{-\frac{1}{2}\left ( \frac{\ln x - \mu}{\sigma} \right )^{2}}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the LogNormal distribution
        sigma : numpy array or scalar
            The scale parameter for the LogNormal distribution

        Returns
        -------

        df : scalar or numpy array 
            The value(s) of the density function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogNormal
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> LogNormal.df(x, 3, 4)
        array([0.07528436, 0.04222769, 0.02969364, 0.02298522, 0.01877747])
        """
        return 1./x * norm.pdf(np.log(x), mu, sigma)

    def hf(self, x, mu, sigma):
        r"""

        Instantaneous hazard rate for the LogNormal Distribution:

        .. math::
            h(x) = \frac{f(x)}{R(x)}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the LogNormal distribution
        sigma : numpy array or scalar
            The scale parameter for the LogNormal distribution

        Returns
        -------

        hf : scalar or numpy array 
            The value(s) of the instantaneous hazard rate at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogNormal
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> LogNormal.hf(x, 3, 4)
        array([0.09734551, 0.05881839, 0.04349249, 0.03500202, 0.02952687])
        """
        return self.df(x, mu, sigma) / self.sf(x, mu, sigma)

    def Hf(self, x, mu, sigma):
        r"""

        Cumulative hazard rate for the LogNormal Distribution:

        .. math::
            H(x) = -\ln \left ( R(x) \right )

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        mu : numpy array or scalar
            The location parameter for the LogNormal distribution
        sigma : numpy array or scalar
            The scale parameter for the LogNormal distribution

        Returns
        -------

        Hf : scalar or numpy array 
            The value(s) of the cumulative hazard rate at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogNormal
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> LogNormal.Hf(x, 3, 4)
        array([0.25699427, 0.33137848, 0.3816556 , 0.4205543 , 0.45264333])
        """
        return -np.log(self.sf(x, mu, sigma))

    def qf(self, p, mu, sigma):
        r"""

        Quantile function for the LogNormal Distribution:

        .. math::
            q(p) = e^{\mu + \sigma \Phi^{-1} \left( p \right )}

        Parameters
        ----------

        p : numpy array or scalar
            The percentiles at which the quantile will be calculated 
        mu : numpy array or scalar
            The location parameter for the LogNormal distribution
        sigma : numpy array or scalar
            The scale parameter for the LogNormal distribution

        Returns
        -------

        q : scalar or numpy array 
            The quantiles for the LogNormal distribution at each value p.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogNormal
        >>> p = np.array([.1, .2, .3, .4, .5])
        >>> LogNormal.qf(p, 3, 4)
        array([ 0.11928899,  0.69316658,  2.46550819,  7.29078766, 20.08553692])
        """
        return np.exp(scipy_norm.ppf(p, mu, sigma))

    def mean(self, mu, sigma):
        r"""

        Quantile function for the LogNormal Distribution:

        .. math::
            E = e^{\mu + \frac{\sigma^2}{2}}

        Parameters
        ----------

        p : numpy array or scalar
            The percentiles at which the quantile will be calculated 
        mu : numpy array or scalar
            The location parameter for the LogNormal distribution
        sigma : numpy array or scalar
            The scale parameter for the LogNormal distribution

        Returns
        -------

        q : scalar or numpy array 
            The quantiles for the LogNormal distribution at each value p.

        Examples
        --------
        >>> from surpyval import LogNormal
        >>> LogNormal.mean(3, 4)
        59874.14171519782
        """
        return np.exp(mu + (sigma**2)/2)

    def moment(self, n, mu, sigma):
        r"""

        n-th (non central) moment of the LogNormal distribution

        .. math::
            E = ... complicated.

        Parameters
        ----------

        n : integer or numpy array of integers
            The ordinal of the moment to calculate
        mu : numpy array or scalar
            The location parameter for the LogNormal distribution
        sigma : numpy array or scalar
            The scale parameter for the LogNormal distribution

        Returns
        -------

        moment : scalar or numpy array 
            The moment(s) of the LogNormal distribution

        Examples
        --------
        >>> from surpyval import LogNormal
        >>> LogNormal.moment(2, 3, 4)
        3.1855931757113756e+16
        """
        return np.exp(n*mu + (n**2 * sigma**2)/2)

    def random(self, size, mu, sigma):
        r"""

        Draws random samples from the distribution in shape `size`

        Parameters
        ----------

        size : integer or tuple of positive integers
            Shape or size of the random draw
        mu : numpy array or scalar
            The location parameter for the LogNormal distribution
        sigma : numpy array or scalar
            The scale parameter for the LogNormal distribution

        Returns
        -------

        random : scalar or numpy array 
            Random values drawn from the distribution in shape `size`

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogNormal
        >>> LogNormal.random(10, 3, 4)
        array([1.74605298e+00, 1.90729963e+02, 1.90090366e+03, 2.59154042e-02,
               3.71460694e-02, 3.38580771e+03, 7.58826512e+04, 7.23252303e+00,
               1.21226718e+03, 4.15054624e+00])
        >>> LogNormal.random((5, 5), 3, 4)
        array([[4.59689256e+00, 2.91472936e-01, 4.66833783e+02, 9.88539048e+01,
                3.88094471e+01],
               [7.10705735e-01, 5.00788529e-02, 2.49032431e+01, 2.19196376e+01,
                2.05043988e+02],
               [1.32193999e+03, 7.38943238e-01, 5.16503535e-01, 9.09249819e+02,
                2.69407879e+03],
               [7.29473033e+00, 5.68246498e+03, 1.74464896e+00, 1.26043004e+00,
                3.84009666e+03],
               [1.47997384e+00, 2.21809242e+02, 1.32564564e+02, 8.06883052e-02,
                1.05118538e+02]])
        """
        return np.exp(para.Normal.random(size, mu, sigma))

    def mpp_x_transform(self, x, gamma=0):
        return np.log(x - gamma)

    def mpp_y_transform(self, y, *params):
        return para.Normal.qf(y, 0, 1)

    def mpp_inv_y_transform(self, y, *params):
        return para.Normal.ff(y, 0, 1)

    def unpack_rr(self, params, rr):
        if rr == 'y':
            sigma, mu = params
            mu = -mu/sigma
            sigma = 1./sigma
        elif rr == 'x':
            sigma, mu = params
        return mu, sigma

    def var_z(self, x, mu, sigma, cv_matrix):
        z_hat = (x - mu)/sigma
        var_z = (1./sigma)**2 * (cv_matrix[0, 0] + z_hat**2 * cv_matrix[1, 1] + 
            2 * z_hat * cv_matrix[0, 1])
        return var_z

    def z_cb(self, x, mu, sigma, cv_matrix, alpha_ci=0.05):
        z_hat = (x - mu)/sigma
        var_z = self.var_z(x, mu, sigma, cv_matrix)
        bounds = z_hat + np.array([1., -1.]).reshape(2, 1) * z(alpha_ci/2) * np.sqrt(var_z)
        return bounds

    # def R_cb(self, x, mu, sigma, cv_matrix, alpha_ci=0.05):
        # t = np.log(x)
        # return para.Normal.sf(self.z_cb(t, mu, sigma, cv_matrix, alpha_ci=alpha_ci), 0, 1).T

    def _mom(self, x):
        norm_mod = para.Normal.fit(np.log(x), how='MOM')
        mu, sigma = norm_mod.params
        return mu, sigma

LogNormal = LogNormal_('LogNormal')
Galton = LogNormal_('Galton')


from surpyval import np
from autograd import jacobian
from scipy.stats import uniform
from autograd.scipy.special import gamma as agamma
from scipy.special import gammaincinv
from autograd_gamma import gammainc as agammainc
from scipy.special import ndtri as z
from scipy.optimize import minimize
from scipy.stats import pearsonr
import warnings
from surpyval import nonparametric as nonp
from surpyval.parametric.parametric_fitter import ParametricFitter


class Gamma_(ParametricFitter):
    r"""

    Class used to generate the Gamma class.

    .. code:: python

        from surpyval import Gamma

    """
    def __init__(self, name):
        self.name = name
        self.k = 2
        self.bounds = ((0, None), (0, None),)
        self.support = (0, np.inf)
        self.plot_x_scale = 'linear'
        self.y_ticks = [0.0001, 0.0002, 0.0003, 0.001, 0.002,
                        0.003, 0.005, 0.01, 0.02, 0.03, 0.05,
                        0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8,
                        0.9, 0.95, 0.99, 0.999, 0.9999]
        self.param_names = ['alpha', 'beta']
        self.param_map = {
            'alpha': 0,
            'beta': 1
        }

    def _parameter_initialiser(self, x, c=None, n=None, offset=False):
        # These equations are truly magical
        if offset:
            s = np.log(x.sum()/len(x)) - np.log(x).sum()/len(x)
            alpha = (3 - s + np.sqrt((s - 3)**2 + 24*s)) / (12*s)
            beta = x.sum()/(len(x)*alpha)
            return alpha, 1./beta, np.min(x) - 1
        else:
            s = np.log(x.sum()/len(x)) - np.log(x).sum()/len(x)
            alpha = (3 - s + np.sqrt((s - 3)**2 + 24*s)) / (12*s)
            beta = x.sum()/(len(x)*alpha)
            return alpha, 1./beta

    def sf(self, x, alpha, beta):
        r"""

        Surival (or Reliability) function for the Gamma Distribution:

        .. math::
            R(x) = 1 - \frac{\gamma \left ( \alpha, \beta x \right )
            }{\Gamma \left ( \alpha \right )}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated
        alpha : numpy array or scalar
            The shape parameter for the Gamma distribution
        beta : numpy array or scalar
            The scale parameter for the Gamma distribution


        Returns
        -------

        sf : scalar or numpy array
            The value(s) for the survival function at each x

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gamma
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Gamma.sf(x, 3, 2)
        array([0.67667642, 0.23810331, 0.0619688 , 0.01375397, 0.0027694 ])
        """
        return 1 - self.ff(x, alpha, beta)

    def cs(self, x, X, alpha, beta):
        r"""

        Conditional survival function for the Gamma Distribution:

        .. math::
            R(x) = e^{-\lambda x}

        Parameters
        ----------

        x : numpy array or scalar
            The value(s) at which the function will be calculated
        X : numpy array or scalar
            The value(s) at which each value(s) in x was known to have survived
        alpha : numpy array or scalar
            The shape parameter for the Gamma distribution
        beta : numpy array or scalar
            The scale parameter for the Gamma distribution

        Returns
        -------

        cs : scalar or numpy array
            the conditional survival probability.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gamma
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Gamma.cs(x, 5, 3, 4)
        array([2.59402488e-02, 6.39048747e-04, 1.51519143e-05, 3.48776510e-07,
               7.79933496e-09])
        """
        return self.sf(x + X, alpha, beta) / self.sf(X, alpha, beta)

    def ff(self, x, alpha, beta):
        r"""

        CDF (or unreliability or failure) function for the Gamma Distribution:

        .. math::
            F(x) = \frac{\gamma \left ( \alpha, \beta x \right )}
            {\Gamma \left ( \alpha \right )}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated
        alpha : numpy array or scalar
            The shape parameter for the Gamma distribution
        beta : numpy array or scalar
            The scale parameter for the Gamma distribution


        Returns
        -------

        ff : scalar or numpy array
            The value(s) for the CDF at each x

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gamma
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Gamma.ff(x, 3, 2)
        array([0.32332358, 0.76189669, 0.9380312 , 0.98624603, 0.9972306 ])
        """
        return agammainc(alpha, beta * x)

    def df(self, x, alpha, beta):
        r"""

        Density function for the Gamma Distribution:

        .. math::
            f(x) = \frac{\beta^{\alpha }}{\Gamma \left ( \alpha \right )}
            x^{\alpha - 1}e^{-\beta x}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated
        alpha : numpy array or scalar
            The shape parameter for the Gamma distribution
        beta : numpy array or scalar
            The scale parameter for the Gamma distribution


        Returns
        -------

        df : scalar or numpy array
            The density of the distribution at each x

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gamma
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Gamma.df(x, 3, 2)
        array([0.54134113, 0.29305022, 0.08923508, 0.02146961, 0.00453999])
        """
        return (((beta ** alpha) * x ** (alpha - 1)
                 * np.exp(-(x * beta)) / (agamma(alpha))))

    def hf(self, x, alpha, beta):
        r"""

        Instantaneous hazard rate for the Gamma Distribution:

        .. math::
            h(x) = \frac{\frac{\beta^{\alpha }}{\Gamma \left ( \alpha \right )
            }x^{\alpha - 1}e^{-\beta x}}{1 - \frac{\gamma \left ( \alpha, \beta
            x \right )}{\Gamma \left ( \alpha \right )}}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated
        alpha : numpy array or scalar
            The shape parameter for the Gamma distribution
        beta : numpy array or scalar
            The scale parameter for the Gamma distribution


        Returns
        -------

        Hf : scalar or numpy array
            The instantaneous hazard rate of the distribution at each x

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gamma
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Gamma.hf(x, 3, 2)
        array([0.8       , 1.23076923, 1.44      , 1.56097561, 1.63934426])
        """
        return self.df(x, alpha, beta) / self.sf(x, alpha, beta)

    def Hf(self, x, alpha, beta):
        r"""

        Cumulative hazard rate for the Gamma Distribution:

        .. math::
            H(x) = -\ln(1 - \frac{\gamma \left ( \alpha, \beta x \right )}
            {\Gamma \left ( \alpha \right )})

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated
        alpha : numpy array or scalar
            The shape parameter for the Gamma distribution
        beta : numpy array or scalar
            The scale parameter for the Gamma distribution


        Returns
        -------

        Hf : scalar or numpy array
            The cumulative hazard rate of the distribution at each x

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gamma
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Gamma.Hf(x, 3, 2)
        array([0.39056209, 1.43505064, 2.78112418, 4.28642793, 5.88912614])
        """
        return -np.log(self.sf(x, alpha, beta))

    def qf(self, p, alpha, beta):
        r"""

        Quantile function for the Gamma Distribution:

        .. math::
            q(p) = \frac{-\ln\left ( p \right )}{\lambda}

        Parameters
        ----------

        p : numpy array or scalar
            The percentiles at which the quantile will be calculated
        alpha : numpy array or scalar
            The shape parameter for the Gamma distribution
        beta : numpy array or scalar
            The scale parameter for the Gamma distribution

        Returns
        -------

        q : scalar or numpy array
            The quantiles for the Gamma distribution at each value p.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gamma
        >>> p = np.array([.1, .2, .3, .4, .5])
        >>> Gamma.qf(p, 3, 4)
        array([0.27551633, 0.38376105, 0.47844395, 0.57126923, 0.66851508])
        """
        return gammaincinv(alpha, p) / beta

    def mean(self, alpha, beta):
        r"""

        Calculates the mean of the Gamma distribution with given parameters.

        .. math::
            E = \frac{\alpha}{\beta}

        Parameters
        ----------

        alpha : numpy array or scalar
            The shape parameter for the Gamma distribution
        beta : numpy array or scalar
            The scale parameter for the Gamma distribution

        Returns
        -------

        mean : scalar or numpy array
            The mean(s) of the Gamma distribution

        Examples
        --------
        >>> from surpyval import Gamma
        >>> Gamma.mean(3, 4)
        0.75
        """
        return alpha / beta

    def moment(self, n, alpha, beta):
        r"""

        Calculates the n-th moment of the Gamma distribution with
        given parameters.

        .. math::
            E = \frac{\Gamma \left ( n + \alpha \right )}{\beta^{n}\Gamma
            \left ( \alpha \right )}

        Parameters
        ----------

        n : integer or numpy array of integers
            The ordinal of the moment to calculate
        alpha : numpy array or scalar
            The shape parameter for the Gamma distribution
        beta : numpy array or scalar
            The scale parameter for the Gamma distribution

        Returns
        -------

        mean : scalar or numpy array
            The moment(s) of the Gamma distribution

        Examples
        --------
        >>> from surpyval import Gamma
        >>> Gamma.moment(3, 3, 4)
        0.9375
        """
        return agamma(n + alpha) / (beta**n * agamma(alpha))

    def random(self, size, alpha, beta):
        r"""

        Draws random samples from the Gamma distribution in shape `size`

        Parameters
        ----------

        alpha : numpy array or scalar
            The shape parameter for the Gamma distribution
        beta : numpy array or scalar
            The scale parameter for the Gamma distribution

        Returns
        -------

        random : scalar or numpy array
            Random values drawn from the distribution in shape `size`

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gamma
        >>> Gamma.random(10, 3, 4)
        array([0.22856155, 1.69542468, 0.70894789, 0.75552168, 0.76634128,
               0.58624638, 1.03288812, 0.85768925, 0.75071764, 0.91979151])
        >>> Gamma.random((5, 5), 3, 4)
        array([[0.55481976, 1.02867642, 1.25525161, 0.5141736 , 0.7227451 ],
               [1.59192864, 1.22897457, 0.80820007, 0.39872068, 0.53656654],
               [0.80703614, 0.75406597, 0.87307426, 1.88748737, 0.78115455],
               [1.3233755 , 0.29908068, 1.88304902, 2.65690385, 0.51018073],
               [0.36070265, 0.48834586, 0.45623895, 0.30104303, 0.49942908]])
        """
        U = uniform.rvs(size=size)
        return self.qf(U, alpha, beta)

    def mpp_y_transform(self, y, *params):
        alpha = params[0]
        return gammaincinv(alpha, y)

    def mpp_inv_y_transform(self, y, *params):
        alpha = params[0]
        return agammainc(alpha, y)

    def mpp_x_transform(self, x, gamma=0):
        return x - gamma

    def mpp(self, x, c=None, n=None, heuristic="Nelson-Aalen",
            rr='y', on_d_is_0=False, offset=False):

        x_pp, r, d, F = nonp.plotting_positions(x,
                                                c=c,
                                                n=n,
                                                heuristic=heuristic)

        results = {}

        if on_d_is_0:
            pass
        else:
            F = F[d > 0]
            x_pp = x_pp[d > 0]

        if (F == 1).any():
            mask = F != 1
            warnings.warn("Some heuristic values for CDF = 1 have been "
                          + "encountered in plotting points and have been "
                          + "ignored.", stacklevel=2)
            F = F[mask]
            x_pp = x_pp[mask]

        init = self._parameter_initialiser(x_pp, c, n)

        mask = np.isfinite(F)
        if mask.any():
            warnings.warn("Some Infinite values encountered in plotting "
                          + "points and have been ignored.", stacklevel=2)
            F = F[mask]
            x_pp = x_pp[mask]

        if offset:

            def fun(a):
                return -pearsonr(x_pp, self.mpp_y_transform(F, a, 1.))[0]

            res = minimize(fun, init[0], bounds=((0, None),))
            alpha = res.x[0]

            y_pp = self.mpp_y_transform(F, alpha)

            if rr == 'y':
                params = np.polyfit(x_pp, y_pp, 1)
                beta = params[0]
                gamma = -params[1]/beta
            elif rr == 'x':
                params = np.polyfit(y_pp, x_pp, 1)
                beta = 1./params[0]
                gamma = -params[1]/beta

            results['gamma'] = gamma
            results['params'] = [alpha, beta]

            return results
        else:
            if rr == 'y':
                x_pp = x_pp[:, np.newaxis]

                def fun(alpha):
                    y_pp = self.mpp_y_transform(F, alpha, 1.)
                    return np.linalg.lstsq(x_pp, y_pp)[1]

                res = minimize(fun, init[0], bounds=((0, None),))
                alpha = res.x[0]
                y_pp = self.mpp_y_transform(F, alpha, 1.)
                beta, residuals, _, _ = np.linalg.lstsq(x_pp, y_pp)
                beta = beta[0]
            else:
                def fun(a):
                    y = self.mpp_y_transform(F, a, 1.)[:, np.newaxis]
                    return np.linalg.lstsq(y, x)[1]

                res = minimize(fun, init[0], bounds=((0, None),))
                alpha = res.x[0]
                beta = np.linalg.lstsq(self.mpp_y_transform(F,
                                                            alpha,
                                                            1.)[:, np.newaxis],
                                       x)[0]
                beta = 1./beta

            results['params'] = [alpha, beta]

            return results

    def var_R(self, dR, cv_matrix):
        dr_dalpha = dR[:, 0]
        dr_dbeta = dR[:, 1]
        var_r = (dr_dalpha**2 * cv_matrix[0, 0] +
                 dr_dbeta**2 * cv_matrix[1, 1] +
                 2 * dr_dalpha * dr_dbeta * cv_matrix[0, 1])
        return var_r

    def R_cb_(self, x, alpha, beta, cv_matrix, alpha_ci=0.05):
        R_hat = self.sf(x, alpha, beta)

        def dR_f(t):
            return self.sf(*t)
        jac = jacobian(dR_f)
        x_ = np.array(x)
        if x_.size == 1:
            dR = jac(np.array((x_, alpha, beta))[1::])
            dR = dR.reshape(1, 2)
        else:
            out = []
            for xx in x_:
                out.append(jac(np.array((xx, alpha, beta)))[1::])
            dR = np.array(out)
        K = z(alpha_ci/2)
        exponent = (K * np.array([-1, 1]).reshape(2, 1)
                    * np.sqrt(self.var_R(dR, cv_matrix)))
        exponent = exponent/(R_hat*(1 - R_hat))
        R_cb = R_hat / (R_hat + (1 - R_hat) * np.exp(exponent))
        return R_cb.T


Gamma = Gamma_('Gamma')



from .plotting_positions import plotting_positions
from .turnbull import turnbull
from .filliben import filliben
from .kaplan_meier import kaplan_meier
from .nelson_aalen import nelson_aalen
from .fleming_harrington import fleming_harrington
from .success_run import success_run
from .rank_adjust import rank_adjust

from .nonparametric import NonParametric
from .kaplan_meier import KaplanMeier
from .nelson_aalen import NelsonAalen
from .fleming_harrington import FlemingHarrington
from .turnbull import Turnbull

FIT_FUNCS = {
    'Nelson-Aalen': nelson_aalen,
    'Kaplan-Meier': kaplan_meier,
    'Fleming-Harrington': fleming_harrington,
    'Turnbull': turnbull
}

PLOTTING_METHODS = ["Blom", "Median", "ECDF", "Modal", "Midpoint",
                    "Mean", "Weibull", "Benard", "Beard",
                    "Hazen", "Gringorten", "None", "Tukey",
                    "DPW", "Fleming-Harrington", "Kaplan-Meier",
                    "Nelson-Aalen", "Filliben", "Larsen", "Turnbull"]



"""
This code was created for and sponsored by Cartiga (www.cartiga.com). 
Cartiga makes no representations or warranties in connection with the code 
and waives any and all liability in connection therewith. Your use of the 
code constitutes acceptance of these terms.

Copyright 2022 Cartiga LLC
"""

import numpy as np
from surpyval import CoxPH
from surpyval.utils import _get_idx
import numpy_indexed as npi
from surpyval.utils import validate_fine_gray_inputs, _scale
from scipy.optimize import minimize
# from numba import njit
# from scipy.optimize import root

# TODO: Improve this so that it uses root and the jac and hess.
# TODO: Implement it with time-varying covariates

class FineGray_():

    def create_breslow_ll_jac_hess(self, x, Z, c, n, e, mask,
                                   with_jac=True,
                                   with_hess=True):
        """
        The reference used to compute the jacobian and hessian
        was https://mathweb.ucsd.edu/~rxu/math284/slect5.pdf
        """
        # TODO: Incorporate left-truncation... somehow.
        # left truncation allows for implementation of time-varying covariates

        
        x_e, c_e, n_e = [arr[mask] for arr in (x, c, n)]
        Z_e = Z[mask, :]

        x_fg, c_fg, n_fg = [arr[~mask] for arr in (x, c, n)]
        Z_fg = Z[~mask, :]

        gb_x = npi.group_by(x_e)

        n_d_x = np.where(c_e == 0, n_e, 0)
        x_e_unique, n_d = gb_x.sum(n_d_x)
        n_d = n_d.reshape(-1, 1)
        n_e = n_e.reshape(-1, 1)
        
        # Create the log_like function for the data
        def log_like(beta):
            beta_z = Z_e @ beta
            di_beta_z = gb_x.sum(n_d_x * beta_z)[1].reshape(-1, 1)

            e_beta_z = np.exp(beta_z).reshape(-1, 1)
            Ri = gb_x.sum(n_e * e_beta_z)[1]
            Ri = Ri[::-1].cumsum()[::-1].reshape(-1, 1)
            # Add fine gray risk set
            Ri_e = (n_fg * np.exp(Z_fg @ beta))

            Ri += Ri_e.sum()
            Ri = np.log(Ri)
            
            Ri = n_d * Ri
            
            like = di_beta_z - Ri
            
            return -like.sum()

        if with_jac:
            S_d = gb_x.sum(n_d_x.reshape(-1, 1) * Z)[1]

            def jac(beta):
                e_beta_z = np.exp(Z @ beta).reshape(-1, 1)
                z_e_beta_z = Z * e_beta_z

                Ri = gb_x.sum(n * e_beta_z)[1]
                Ri = Ri[::-1].cumsum(axis=0)[::-1]
                Ri += Ri_e

                ZRi = gb_x.sum(n * z_e_beta_z)[1]
                ZRi = ZRi[::-1].cumsum(axis=0)[::-1]
                ZRi += ZRi_e

                EZ = ZRi / Ri
                
                EZ = n_d * EZ

                out = (S_d - EZ)
                return -out.sum(axis=0)
        else:
            jac=None

        
        if with_hess:
            def hess(beta):
                # denominator for both terms
                e_beta_z = np.exp(Z @ beta).reshape(-1, 1)
                denom = at_risk_beta_Z(e_beta_z, n, gb_x)

                # calc term 1
                Z2 = np.einsum('ij, ik-> ijk', Z, Z)
                
                term_1 = np.einsum('ijk,ij->ijk', Z2, n.reshape(-1, 1) * e_beta_z)
                term_1 = gb_x.sum(term_1)[1]
                term_1 = term_1[::-1].cumsum(axis=0)[::-1]
                term_1 = np.einsum('ijk,ij->ijk', term_1, 1./denom)
                
                # calc term 2
                z_beta_z = Z * e_beta_z
                term_2 = at_risk_beta_Z(z_beta_z, n, gb_x)
                term_2 = term_2 / denom
                term_2 = np.einsum('ij, ik-> ijk', term_2, term_2) 
                
                # Compute the Hessian matrix
                hess_matrix = term_1 - term_2
                hess_matrix = np.einsum('ijk,i->ijk', hess_matrix, n_d.flatten())
                hess_matrix = hess_matrix.sum(axis=0)
                return hess_matrix
        else:
            hess = None

        return log_like, jac, hess

    def fit(self, x, Z, e, c=None, n=None):

        x, Z, e, c, n = validate_fine_gray_inputs(x, Z, e, c, n)

        unique_e = list(set(e))
        if None in unique_e:
            unique_e.remove(None)

        # Best initial assumption is to assume there is no risk
        beta_init = np.zeros(Z.shape[1])

        results = []
        for i, event in enumerate(unique_e):
            mask = e == event
            neg_ll, jac, hess = self.create_breslow_ll_jac_hess(x, Z, c, n, e, mask,
                                                                with_jac=False,
                                                                with_hess=False)

            # Have found that root finding is faster than minimization!
            # res = root(jac, beta_init, jac=hess, tol=tol)
            # fun = lambda beta: -self.partial_log_like(beta, x, c_e, n, Z, e, event)
            res = minimize(neg_ll, beta_init)
            
            results.append(res)

        return results, unique_e

FineGray = FineGray_()



from surpyval import np

def pass_through(x):
    return x


def adj_relu(x):
    return np.where(x >= 0, x + 1, np.exp(x))


def inv_adj_relu(x):
    return np.where(x >= 1, x - 1, np.log(x))


def rev_adj_relu(x):
    return -np.where(x >= 0, x + 1, np.exp(x))


def inv_rev_adj_relu(x):
    return np.where(x < -1, -x - 1, np.log(-x))


def bounds_convert(x, bounds):
    funcs = []
    inv_f = []
    for i, (lower, upper) in enumerate(bounds):
        def add_to_funcs(low, upp, i):
            if (low is None) and (upp is None):
                funcs.append(lambda x: pass_through(x))
                inv_f.append(lambda x: pass_through(x))
            elif (low == 0) and (upp == 1):
                D = 10
                funcs.append(lambda x: D * np.arctanh((2 * x) - 1))
                inv_f.append(lambda x: (np.tanh(x / D) + 1) / 2)
            elif (upp is None):
                funcs.append(lambda x: (inv_adj_relu(x - low)))
                inv_f.append(lambda x: (adj_relu(x) + low))
            elif (low is None):
                funcs.append(lambda x: inv_rev_adj_relu(x - np.copy(upp)))
                inv_f.append(lambda x: np.copy(upp) + rev_adj_relu(x))
            else:
                funcs.append(lambda x: pass_through(x))
                inv_f.append(lambda x: pass_through(x))
        add_to_funcs(lower, upper, i)

    def transform(params):
        return np.array([f(p) for p, f in zip(params, funcs)])

    def inv_trans(params):
        return np.array([f(p) for p, f in zip(params, inv_f)])

    return transform, inv_trans, funcs, inv_f


def fix_idx_and_function(fixed, param_map, funcs):
    n_params = len(param_map)
    if fixed is not None:
        """
        Record to the model that parameters were fixed
        """
        fixed_idx = [param_map[x] for x in fixed.keys()]
        not_fixed = [x for x in range(n_params) if x not in fixed_idx]
        not_fixed = np.array(not_fixed)

        def constraints(p):
            params = [0] * (n_params)
            for k, v in fixed.items():
                params[param_map[k]] = funcs[param_map[k]](v)
            for i, v in zip(not_fixed, p):
                params[i] = v
            return np.array(params)

        const = constraints
    else:

        def const(x):
            return x

        fixed_idx = []
        not_fixed = np.array([x for x in range(n_params)])

    return const, fixed_idx, not_fixed



from autograd import numpy as np

class DualExponential_():
    def __init__(self):
        self.name = 'DualExponential'
        self.phi_param_map = {'a' : 0, 'b' : 1, 'c' : 2}
        self.phi_bounds = ((None, None), (None, None), (0, None))

    def phi(self, X, *params):
        X = np.atleast_2d(X)
        X1 = X[:, 0]
        X2 = X[:, 1]
        a = params[0]
        b = params[1]
        c = params[2]
        return c * np.exp(a/X1 + b/X2)

    def phi_init(self, life, X):
        A = np.atleast_2d(X)
        A = 1./np.hstack([np.ones(X.shape[0]).reshape(-1, 1), X])
        y = np.log(life)
        c, a, b = np.linalg.lstsq(A, y, rcond=None)[0]
        return [a, b, np.exp(c)]

DualExponential = DualExponential_()


from surpyval import np
from scipy.stats import uniform
from numpy import euler_gamma
from scipy.special import gamma as gamma_func
from scipy.special import ndtri as z

from surpyval import xcn_handler
from surpyval import parametric as para
from surpyval.parametric.parametric_fitter import ParametricFitter

class LogLogistic_(ParametricFitter):
    def __init__(self, name):
        self.name = name
        self.k = 2
        self.bounds = ((0, None), (0, None),)
        self.support = (0, np.inf)
        self.plot_x_scale = 'log'
        self.y_ticks = [0.0001, 0.0002, 0.0003, 0.001, 0.002, 
            0.003, 0.005, 0.01, 0.02, 0.03, 0.05, 
            0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 
            0.9, 0.95, 0.99, 0.999, 0.9999]
        self.param_names = ['alpha', 'beta']
        self.param_map = {
            'alpha' : 0,
            'beta' : 1
        }

    def _parameter_initialiser(self, x, c=None, n=None, t=None, offset=False):
        if offset:
            # return *self.fit(x, c, n, t, how='MPP').params, 1.
            x, c, n = xcn_handler(x, c, n)
            flag = (c == 0).astype(int)
            value_range = np.max(x) - np.min(x)
            gamma_init = np.min(x) - value_range / 10
            return gamma_init, x.sum() / (n * flag).sum(), 2., 1.
        else:
            return self.fit(x, c, n, t, how='MPP').params
            # x, c, n = surpyval.xcn_handler(x, c, n)
            # flag = (c == 0).astype(int)
            # return x.sum() / (n * flag).sum(), 2.

    def sf(self, x, alpha, beta):
        r"""

        Survival (or reliability) function for the LogLogistic Distribution:

        .. math::
            R(x) = 1 - \frac{1}{1 + \left ( x /\alpha \right )^{-\beta}}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the LogLogistic distribution
        beta : numpy array or scalar
            shape parameter for the LogLogistic distribution

        Returns
        -------

        sf : scalar or numpy array 
            The value(s) of the reliability function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogLogistic
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> LogLogistic.sf(x, 3, 4)
        array([0.62245933, 0.5621765 , 0.5       , 0.4378235 , 0.37754067])
        """
        return 1 - self.ff(x, alpha, beta)

    def cs(self, x, X, alpha, beta):
        r"""

        Conditional survival function for the LogLogistic Distribution:

        .. math::
            R(x, X) = \frac{R(x + X)}{R(X)}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated
        X : numpy array or scalar
            The value(s) at which each value(s) in x was known to have survived
        alpha : numpy array or scalar
            scale parameter for the LogLogistic distribution
        beta : numpy array or scalar
            shape parameter for the LogLogistic distribution

        Returns
        -------

        cs : scalar or numpy array 
            The value(s) of the conditional survival function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogLogistic
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> LogLogistic.cs(x, 5, 3, 4)
        array([0.51270879, 0.28444803, 0.16902083, 0.10629329, 0.07003273])
        """
        return self.sf(x + X, alpha, beta) / self.sf(X, alpha, beta)

    def ff(self, x, alpha, beta):
        r"""

        Failure (CDF or unreliability) function for the LogLogistic Distribution:

        .. math::
            F(x) = \frac{1}{1 + \left ( x /\alpha \right )^{-\beta}}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the LogLogistic distribution
        beta : numpy array or scalar
            shape parameter for the LogLogistic distribution

        Returns
        -------

        ff : scalar or numpy array 
            The value(s) of the failure function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogLogistic
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> LogLogistic.ff(x, 3, 4)
        array([0.01219512, 0.16494845, 0.5       , 0.75964392, 0.88526912])
        """
        return 1. / (1 + (x/alpha)**-beta)

    def df(self, x, alpha, beta):
        r"""

        Density function for the LogLogistic Distribution:

        .. math::
            f(x) = \frac{\left ( \beta / \alpha \right ) \left ( x / \alpha \right )^{\beta - 1}}{\left ( 1 + \left ( x / \alpha \right )^{-\beta} \right )^2}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the LogLogistic distribution
        beta : numpy array or scalar
            shape parameter for the LogLogistic distribution

        Returns
        -------

        df : scalar or numpy array 
            The value(s) of the failure function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogLogistic
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> LogLogistic.df(x, 3, 4)
        array([0.0481856 , 0.27548092, 0.33333333, 0.18258504, 0.08125416])
        """
        return ((beta/alpha)*(x/alpha)**(beta-1.))/((1. + (x/alpha)**beta)**2.)

    def hf(self, x, alpha, beta):
        r"""

        Instantaneous hazard rate for the LogLogistic Distribution:

        .. math::
            h(x) = \frac{f(x)}{R(x)}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the LogLogistic distribution
        beta : numpy array or scalar
            shape parameter for the LogLogistic distribution

        Returns
        -------

        hf : scalar or numpy array 
            The value(s) of the instantaneous hazard rate at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogLogistic
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> LogLogistic.hf(x, 3, 4)
        array([0.04878049, 0.32989691, 0.66666667, 0.75964392, 0.7082153 ])
        """
        return self.df(x, alpha, beta) / self.sf(x, alpha, beta)

    def Hf(self, x, alpha, beta):
        r"""

        Cumulative hazard rate for the LogLogistic Distribution:

        .. math::
            H(x) = -\ln \left ( R(x) \right )

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the LogLogistic distribution
        beta : numpy array or scalar
            shape parameter for the LogLogistic distribution

        Returns
        -------

        Hf : scalar or numpy array 
            The value(s) of the cumulative hazard rate at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogLogistic
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> LogLogistic.Hf(x, 3, 4)
        array([0.01227009, 0.18026182, 0.69314718, 1.42563378, 2.16516608])
        """
        return -np.log(self.sf(x, alpha, beta))

    def qf(self, p, alpha, beta):
        r"""

        Quantile function for the LogLogistic distribution:

        .. math::
            q(p) = \alpha \left ( \frac{p}{1 - p} \right )^{\frac{1}{\beta}}

        Parameters
        ----------

        p : numpy array or scalar
            The percentiles at which the quantile will be calculated
        alpha : numpy array or scalar
            scale parameter for the LogLogistic distribution
        beta : numpy array or scalar
            shape parameter for the LogLogistic distribution

        Returns
        -------

        q : scalar or numpy array 
            The quantiles for the LogLogistic distribution at each value p

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogLogistic
        >>> p = np.array([.1, .2, .3, .4, .5])
        >>> LogLogistic.qf(p, 3, 4)
        array([1.73205081, 2.12132034, 2.42732013, 2.71080601, 3.        ])
        """
        return alpha * (p/(1 - p))**(1./beta)

    def mean(self, alpha, beta):
        r"""

        Mean of the LogLogistic distribution

        .. math::
            E = \frac{\alpha \pi / \beta}{sin \left ( \pi / \beta \right )}

        Parameters
        ----------

        alpha : numpy array or scalar
            scale parameter for the LogLogistic distribution
        beta : numpy array or scalar
            shape parameter for the LogLogistic distribution

        Returns
        -------

        mean : scalar or numpy array 
            The mean(s) of the LogLogistic distribution

        Examples
        --------
        >>> from surpyval import LogLogistic
        >>> LogLogistic.mean(3, 4)
        3
        """
        if beta > 1:
            return (alpha * np.pi / beta) / (np.sin(np.pi/beta))
        else:
            return np.nan

    def random(self, size, alpha, beta):
        r"""

        Draws random samples from the distribution in shape `size`

        Parameters
        ----------

        size : integer or tuple of positive integers
            Shape or size of the random draw
        alpha : numpy array or scalar
            scale parameter for the LogLogistic distribution
        beta : numpy array or scalar
            shape parameter for the LogLogistic distribution

        Returns
        -------

        random : scalar or numpy array 
            Random values drawn from the distribution in shape `size`

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import LogLogistic
        >>> LogLogistic.random(10, 3, 4)
        array([4.46072122, 2.1336253 , 2.74159711, 2.90125715, 3.2390347 ,
               5.45223664, 4.28281376, 2.7017541 , 3.023811  , 2.16225601])
        >>> LogLogistic.random((5, 5), 3, 4)
        array([[1.97744499, 4.02823921, 1.95761719, 1.20481591, 3.7166738 ],
               [2.94863864, 3.02609811, 3.30563774, 2.39100075, 3.24937459],
               [3.16102391, 1.77003533, 4.73831093, 0.36936215, 1.41566853],
               [3.88505024, 2.88183095, 2.43977804, 2.62385959, 3.40881857],
               [1.2349273 , 1.83914641, 3.68502568, 6.49834769, 8.62995574]])
        """
        U = uniform.rvs(size=size)
        return self.qf(U, alpha, beta)

    def mpp_x_transform(self, x, gamma=0):
        return np.log(x - gamma)

    def mpp_y_transform(self, y, *params):
        mask = ((y == 0) | (y == 1))
        out = np.zeros_like(y)
        out[~mask] = -np.log(1./y[~mask] - 1)
        out[mask] = np.nan
        return out

    def mpp_inv_y_transform(self, y, *params):
        return 1./(np.exp(-y) + 1)

    def unpack_rr(self, params, rr):
        if rr == 'y':
            beta  = params[0]
            alpha = np.exp(params[1]/-beta)
        elif rr == 'x':
            beta  = 1./params[0]
            alpha = np.exp(params[1] / (beta * params[0]))
        return alpha, beta

LogLogistic = LogLogistic_('LogLogistic')


from .power import InversePower, Power
from .exponential import InverseExponential, Exponential
from .linear import Linear
from .eyring import Eyring, InverseEyring
from .dual_exponential import DualExponential
from .dual_power import DualPower
from .power_exponential import PowerExponential
from .general_log_linear import GeneralLogLinear


from surpyval import np

from surpyval import nonparametric as nonp
from scipy.optimize import minimize

from autograd import jacobian, hessian

from surpyval.utils import xcnt_to_xrd

def mse_fun(params, dist, x, F, inv_trans, const):
    return np.sum(((dist.ff(x, *inv_trans(const(params)))) - F)**2)


def mse(model):
    """
    MSE: Mean Square Error
    This is simply fitting the curve to the best estimate from a non-parametric
    estimate.

    This is slightly different in that it fits it to untransformed data on
    the x and y axis. The MPP method fits the curve to the transformed data.
    This is simply fitting a the CDF sigmoid to the nonparametric estimate.
    """
    dist = model.dist
    x, c, n, t = (model.data['x'], model.data['c'],
                  model.data['n'], model.data['t'])

    const = model.fitting_info['const']
    inv_trans = model.fitting_info['inv_trans']
    init = model.fitting_info['init']

    if (-1 in c) or (2 in c):
        out = nonp.turnbull(x, c, n, t, estimator='Fleming-Harrington')
        F = 1 - out['R']
        x = out['x']
    else:
        x, r, d = xcnt_to_xrd(x, c, n, t)
        R = nonp.fleming_harrington(r, d)
        F = 1 - R
    
    mask = np.isfinite(x)
    F = F[mask]
    x = x[mask]

    jac = jacobian(mse_fun)
    hess = hessian(mse_fun)

    old_err_state = np.seterr(all='ignore')

    res = minimize(mse_fun, init, method='Newton-CG', jac=jac, hess=hess,
                   args=(dist, x, F, inv_trans, const))

    if (res.success is False) or (np.isnan(res.x).any()):
        res = minimize(mse_fun, init, method='BFGS', jac=jac,
                       args=(dist, x, F, inv_trans, const))

    if (res.success is False) or (np.isnan(res.x).any()):
        res = minimize(mse_fun, init,
                       args=(dist, x, F, inv_trans, const))

    results = {}
    results['res'] = res
    results['params'] = inv_trans(const(res.x))
    np.seterr(**old_err_state)
    return results



from autograd import numpy as np
from scipy.stats import t, norm
from surpyval.utils.counting_utils import xicn_to_xrd, handle_xicn

class NonParametricCounting():
    def mcf(self, x, how='step'):
        x = np.atleast_1d(x)
        # Let's not assume we can predict above the highest measurement
        if how == 'step':
            idx = np.searchsorted(self.x, x, side='right') - 1
            mcf = self.mcf_hat[idx]
            mcf[np.where(x < self.x.min())] = 0
            mcf[np.where(x > self.x.max())] = np.nan
            mcf[np.where(x < 0)] = np.nan
            return mcf
        elif how == 'interp':
            mcf = np.hstack([[0], self.mcf_hat])
            x_data = np.hstack([[0], self.x])
            mcf = np.interp(x, x_data, mcf)
            mcf[np.where(x > self.x.max())] = np.nan
            return mcf

    def mcf_cb(self, x, bound='two-sided', how='step', 
               confidence=0.95, bound_type='exp', dist='z'):
        # Greenwoods variance using t-stat. Ref found:
        # http://reliawiki.org/index.php/Non-Parametric_Life_Data_Analysis
        assert bound_type in ['exp', 'normal']
        assert dist in ['t', 'z']
        x = np.atleast_1d(x)
        if bound in ['upper', 'lower']:
            if dist == 't':
                stat = t.ppf(1 - confidence, self.r - 1)
            else:
                stat = norm.ppf(1 - confidence, 0, 1)
            if bound == 'upper' : stat = -stat
        elif bound == 'two-sided':
            if dist == 't':
                stat = t.ppf((1 - confidence)/2, self.r - 1)
            else:
                stat = norm.ppf((1 - confidence)/2, 0, 1)
            stat = np.array([-1, 1]).reshape(2, 1) * stat

        if bound_type == 'exp':
            # Exponential Greenwood confidence
            mcf_cb = self.mcf_hat * np.exp(stat * np.sqrt(self.var)/self.mcf_hat)
        else:
            # Normal Greenwood confidence
            mcf_cb = self.mcf_hat + np.sqrt(self.var * self.mcf_hat**2) * stat
        # Let's not assume we can predict above the highest measurement
        if how == 'step':
            mcf_cb[np.where(x < self.x.min())] = 0
            mcf_cb[np.where(x > self.x.max())] = np.nan
            mcf_cb[np.where(x < 0)] = np.nan
            idx = np.searchsorted(self.x, x, side='right') - 1
            if bound == 'two-sided':
                mcf_cb = mcf_cb[:, idx].T
            else:
                mcf_cb = mcf_cb[idx]
        elif how == 'interp':
            if bound == 'two-sided':
                R1 = np.interp(x, self.x, mcf_cb[0, :])
                R2 = np.interp(x, self.x, mcf_cb[1, :])
                mcf_cb = np.vstack([R1, R2]).T
            else:
                mcf_cb = np.interp(x, self.x, mcf_cb)
            mcf_cb[np.where(x > self.x.max())] = np.nan
        return mcf_cb

    @classmethod
    def fit(cls, x, i=None, c=None, n=None):
        out = cls()

        xicn = handle_xicn(x, i, c, n)
        x, r, d = xicn_to_xrd(*xicn)
        
        out.x = x
        out.r = r
        out.d = d

        out.mcf_hat = np.cumsum(d/r)
        var = 1./r**2 * (d*(1 - 1./r)**2 + (r - d)*(0 - 1./r)**2)
        var = (d > 0).astype(int) * var
        out.var = np.cumsum(var)

        return out


from autograd import numpy as np
from scipy.special import lambertw

class Eyring_():
    def __init__(self):
        self.name = 'Eyring'
        self.phi_param_map = {'a' : 0, 'b' : 1}
        self.phi_bounds = ((None, None), (None, None),)

    def phi(self, X, *params):
        a = params[0]
        c = params[1]
        return (1./X) * np.exp(-(c - a/X))

    def phi_init(self, life, X):
        X = X.flatten()
        a, c =  np.polyfit(1./X, np.log(life) + np.log(X), 1)
        return [a, -c]

Eyring = Eyring_()

class InverseEyring_():
    def __init__(self):
        self.name = 'InverseEyring'
        self.phi_param_map = {'a' : 0, 'c' : 1}
        self.phi_bounds = ((None, None), (None, None),)

    def phi(self, X, *params):
        a = params[0]
        c = params[1]
        return 1./((1./X) * np.exp(-(c - a/X)))

    def phi_init(self, life, X):
        X = X.flatten()
        a, c =  np.polyfit(1./X, np.log(1./life) + np.log(X), 1)
        return [a, -c]

InverseEyring = InverseEyring_()


from surpyval import np
from scipy.stats import uniform
from numpy import euler_gamma
from scipy.special import ndtri as z

from surpyval import nonparametric as nonp
from surpyval import parametric as para
from surpyval.parametric.parametric_fitter import ParametricFitter

class Gumbel_(ParametricFitter):
    def __init__(self, name):
        self.name = name
        self.k = 2
        self.bounds = ((None, None), (0, None),)
        self.support = (-np.inf, np.inf)
        self.plot_x_scale = 'linear'
        self.y_ticks = [0.0001, 0.0002, 0.0003, 0.001, 0.002, 
            0.003, 0.005, 0.01, 0.02, 0.03, 0.05, 
            0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 
            0.9, 0.95, 0.99, 0.999, 0.9999]
        self.param_names = ['mu', 'sigma']
        self.param_map = {
            'mu'    : 0,
            'sigma' : 1
        }

    def _parameter_initialiser(self, x, c=None, n=None):
        if (2 in c) or (-1 in c):
            heuristic = "Turnbull"
        else:
            heuristic = "Fleming-Harrington"
        return self.fit(x, c, n, how='MPP', heuristic=heuristic).params

    def sf(self, x, mu, sigma):
        r"""

        Surival (or Reliability) function for the Gumbel Distribution:

        .. math::
            R(x) = 1 - e^{e^{-\left ( x - \mu \right ) / \sigma}}

        Parameters
        ----------

        x : numpy array or scalar
            The values of the random variables at which the survival function will be calculated 
        mu : numpy array like or scalar
            The location parameter of the distribution
        sigma : numpy array like or scalar
            The scale parameter of the distribution

        Returns
        -------

        sf : scalar or numpy array 
            The scalar value of the survival function of the distribution if a scalar was passed. If an array like object was passed then a numpy array is returned with the value of the survival function at each corresponding value in the input array.


        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gumbel
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Gumbel.sf(x, 3, 2)
        array([0.69220063, 0.54523921, 0.36787944, 0.19229565, 0.06598804])
        """
        return np.exp(-np.exp((x - mu)/sigma))

    def cs(self, x, X, mu, sigma):
        return self.sf(x + X, mu, sigma) / self.sf(X, mu, sigma)

    def ff(self, x, mu, sigma):
        r"""

        CDF (or Failure) function for the Gumbel Distribution:

        .. math::
            F(x) = e^{e^{-\left ( x - \mu \right )/\sigma}}

        Parameters
        ----------

        x : numpy array or scalar
            The values of the random variables at which the survival function will be calculated 
        mu : numpy array like or scalar
            The location parameter of the distribution
        sigma : numpy array like or scalar
            The scale parameter of the distribution

        Returns
        -------

        ff : scalar or numpy array 
            The scalar value of the failure function of the distribution if a scalar was passed. If an array like object was passed then a numpy array is returned with the value of the failure function at each corresponding value in the input array.


        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gumbel
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Gumbel.ff(x, 3, 2)
        array([0.30779937, 0.45476079, 0.63212056, 0.80770435, 0.93401196])
        """
        return 1 - np.exp(-np.exp((x - mu)/sigma))

    def df(self, x, mu, sigma):
        r"""

        Density function (pdf) for the Gumbel Distribution:

        .. math::
            f(x) = \frac{1}{\sigma}e^{\left (\frac{x - \mu}{\sigma} - e^{\frac{x-\mu}{\sigma}} \right)}

        Parameters
        ----------

        x : numpy array or scalar
            The values of the random variables at which the survival function will be calculated 
        mu : numpy array like or scalar
            The location parameter of the distribution
        sigma : numpy array like or scalar
            The scale parameter of the distribution

        Returns
        -------

        df : scalar or numpy array 
            The scalar value of the density of the distribution if a scalar was passed. If an array like object was passed then a numpy array is returned with the value of the density at each corresponding value in the input array.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gumbel
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Gumbel.df(x, 3, 2)
        array([0.12732319, 0.16535215, 0.18393972, 0.15852096, 0.08968704])
        """
        z = (x - mu) / sigma
        return (1/sigma) * np.exp(z - np.exp(z))

    def hf(self, x, mu, sigma):
        r"""

        Instantaneous hazard rate for the Gumbel Distribution:

        .. math::
            h(x) = \frac{1}{\sigma} e^{\frac{x-\mu}{\sigma}}

        Parameters
        ----------

        x : numpy array or scalar
            The values of the random variables at which the survival function will be calculated 
        mu : numpy array like or scalar
            The location parameter of the distribution
        sigma : numpy array like or scalar
            The scale parameter of the distribution

        Returns
        -------

        hf : scalar or numpy array 
            The value(s) for the instantaneous hazard rate for the Gumbel distribution.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gumbel
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Gumbel.hf(x, 3, 2)
        array([0.18393972, 0.30326533, 0.5       , 0.82436064, 1.35914091])
        """
        z = (x - mu) / sigma
        return (1/sigma) * np.exp(z)

    def Hf(self, x, mu, sigma):
        r"""

        Cumulative hazard rate for the Gumbel Distribution:

        .. math::
            H(x) = e^{\frac{x-\mu}{\sigma}}

        Parameters
        ----------

        x : numpy array or scalar
            The values of the random variables at which the survival function will be calculated 
        mu : numpy array like or scalar
            The location parameter of the distribution
        sigma : numpy array like or scalar
            The scale parameter of the distribution

        Returns
        -------

        Hf : scalar or numpy array 
            The value(s) for the cumulative hazard rate for the Gumbel distribution.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gumbel
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Gumbel.Hf(x, 3, 2)
        array([0.36787944, 0.60653066, 1.        , 1.64872127, 2.71828183])
        """
        return np.exp((x - mu)/sigma)

    def qf(self, p, mu, sigma):
        r"""

        Quantile function for the Gumbel Distribution:

        .. math::
            q(p) = \mu + \sigma\ln\left ( -\ln\left ( 1 - p \right ) \right )

        Parameters
        ----------

        p : numpy array or scalar
            The percentiles at which the quantile will be calculated 
        mu : numpy array like or scalar
            The location parameter(s) of the distribution
        sigma : numpy array like or scalar
            The scale parameter(s) of the distribution

        Returns
        -------

        q : scalar or numpy array 
            The quantiles for the Gumbel distribution at each value p.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gumbel
        >>> p = np.array([.1, .2, .3, .4, .5])
        >>> Gumbel.qf(p, 3, 2)
        array([-1.50073465e+00,  1.20026481e-04,  9.38139134e-01,  1.65654602e+00,
        2.26697416e+00])
        """
        return mu + sigma * (np.log(-np.log(1 - p)))

    def mean(self, mu, sigma):
        r"""

        Calculates the mean of the Gumbel distribution with given parameters.

        .. math::
            E = \mu + \sigma\gamma

        Where gamma is the Euler-Mascheroni constant

        Parameters
        ----------

        mu : numpy array like or scalar
            The location parameter(s) of the distribution
        sigma : numpy array like or scalar
            The scale parameter(s) of the distribution

        Returns
        -------

        mean : scalar or numpy array 
            The mean(s) of the Gumbel distribution 

        Examples
        --------
        >>> from surpyval import Gumbel
        >>> Gumbel.mean(3, 2)
        4.1544313298030655
        """
        return mu + sigma * euler_gamma

    def random(self, size, mu, sigma):
        r"""

        Draws random samples from the distribution in shape `size`

        Parameters
        ----------

        size : integer or tuple of positive integers
            Shape or size of the random draw
        mu : numpy array like or scalar
            The location parameter(s) of the distribution
        sigma : numpy array like or scalar
            The scale parameter(s) of the distribution

        Returns
        -------

        random : scalar or numpy array 
            Random values drawn from the distribution in shape `size`

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Gumbel
        >>> Gumbel.random(10, 3, 2)
        array([1.50706388, 3.3098799 , 4.32358009, 2.9914246 , 4.47216839,
               3.56676358, 4.19781514, 4.49123942, 7.29849677, 6.32996653])
        >>> Gumbel.random((5, 5), 3, 2)
        array([[ 5.97265715,  5.89177067,  2.95883424,  2.46315557,  5.15250379],
               [ 2.33808212,  7.42817091,  0.90560051,  8.05897841,  6.30714544],
               [ 6.13076426,  6.31925048,  4.34031705,  3.01309504, -0.70053049],
               [ 5.84888474,  5.95097491,  6.23960618,  6.24830057,  4.89655192],
               [ 6.29507963,  4.21798292,  4.22835474,  5.23521822,  2.76053242]])
        """
        U = uniform.rvs(size=size)
        return self.qf(U, mu, sigma)

    def mpp_x_transform(self, x, gamma=0):
        return x - gamma

    def mpp_y_transform(self, y, *params):
        mask = ((y == 0) | (y == 1))
        out = np.zeros_like(y)
        out[~mask] = np.log(-np.log((1 - y[~mask])))
        out[mask] = np.nan
        return out

    def mpp_inv_y_transform(self, y, *params):
        return 1 - np.exp(-np.exp(y))

    def unpack_rr(self, params, rr):
        if   rr == 'y':
            sigma = 1./params[0]
            mu    = -sigma * params[1]
        elif rr == 'x':
            sigma  = params[0]
            mu = params[1]
        return mu, sigma

    def var_z(self, x, mu, sigma, cv_matrix):
        z_hat = (x - mu)/sigma
        var_z = (1./sigma)**2 * (cv_matrix[0, 0] + z_hat**2 * cv_matrix[1, 1] + 
            2 * z_hat * cv_matrix[0, 1])
        return var_z

    def z_cb(self, x, mu, sigma, cv_matrix, alpha_ci=0.05):
        z_hat = (x - mu)/sigma
        var_z = self.var_z(x, mu, sigma, cv_matrix)
        bounds = z_hat + np.array([1., -1.]).reshape(2, 1) * z(alpha_ci/2) * np.sqrt(var_z)
        return bounds

    # def R_cb(self, x, mu, sigma, cv_matrix, alpha_ci=0.05):
        # return self.sf(self.z_cb(x, mu, sigma, cv_matrix, alpha_ci=alpha_ci), 0, 1).T

Gumbel = Gumbel_('Gumbel')